import unittest
import os
from tests.test_base import BaseAgentTest
from yacana import Task
from yacana.mcp import Mcp
from yacana.tool import ToolType


class TestMcp(BaseAgentTest):
    """Test basic inference capabilities of all agent types."""

    def setUp(self):
        """Clean up agent histories before each test."""
        super().setUp()
        if self.run_ollama:
            self.ollama_agent.history.clean()
        if self.run_openai:
            self.openai_agent.history.clean()
        if self.run_vllm:
            self.vllm_agent.history.clean()
        if self.run_lmstudio:
            self.lmstudio_agent.history.clean()

    def _get_agents(self):
        agents = []
        if self.run_ollama:
            agents.append(("ollama", self.ollama_agent))
        if self.run_openai:
            agents.append(("openai", self.openai_agent))
        if self.run_vllm:
            agents.append(("vllm", self.vllm_agent))
        if self.run_lmstudio:
            agents.append(("lmstudio", self.vllm_agent))
        return agents

    def _reset_agents(self):
        """Reset agent histories after each test."""
        if self.run_ollama:
            self.ollama_agent.history.clean()
        if self.run_openai:
            self.openai_agent.history.clean()
        if self.run_vllm:
            self.vllm_agent.history.clean()
        if self.run_lmstudio:
            self.lmstudio_agent.history.clean()

    def test_connection(self):
        """Test connection to MCP servers and checks that the number of tools matches the expected number for that server."""
        servers = [
            ("deepwiki", "https://mcp.deepwiki.com/mcp", {"read_wiki_structure", "read_wiki_contents", "ask_question"}),
            ("local", "http://localhost:1337/mcp", {"example_tool", "weather"})
        ]
        for name, url, expected_tools in servers:
            mcp = Mcp(url)
            mcp.connect()
            tool_names = {tool.tool_name for tool in mcp.tools}
            self.assertTrue(expected_tools.issubset(tool_names), f"{name} MCP tools mismatch: {tool_names}")
            mcp.disconnect()

    def test_error_handling(self):
        """
        Test for mixing tool execution types YACANA and OPENAI, which should raise an error.
        """
        mcp = Mcp("https://mcp.deepwiki.com/mcp")
        mcp.connect()
        # On mélange les types de tools
        tools = mcp.get_tools_as(ToolType.OPENAI)
        # On force un tool local de type YACANA
        def dummy_tool(x: int): return x
        from yacana.tool import Tool
        local_tool = Tool("dummy", "desc", dummy_tool, tool_type=ToolType.YACANA)
        mixed_tools = tools + [local_tool]
        for _, agent in self._get_agents():
            with self.assertRaises(Exception):
                Task("Test mixing tool types", agent, tools=mixed_tools).solve()
        mcp.disconnect()

    def test_local_tool_override_mcp_tool(self):
        """
        Test that local tools override MCP tools with the same name.
        """
        mcp = Mcp("http://localhost:1337/mcp")
        mcp.connect()
        # MCP fournit "example_tool"
        def example_tool(name: str):
            return {"result": f"local override: {name}"}
        from yacana.tool import Tool
        local_tool = Tool("example_tool", "Override", example_tool, tool_type=ToolType.OPENAI)
        tools = [local_tool] + mcp.get_tools_as(ToolType.OPENAI)
        for _, agent in self._get_agents():
            msg = Task("Call example_tool with name 'yacana'", agent, tools=tools).solve()
            self.assertIn("local override", msg.content)
        mcp.disconnect()

    def test_forget_about_tool(self):
        """
        Test that the forget_tool method works as expected.
        """
        mcp = Mcp("https://mcp.deepwiki.com/mcp")
        mcp.connect()
        self.assertTrue(any(t.tool_name == "read_wiki_structure" for t in mcp.tools))
        mcp.forget_tool("read_wiki_structure")
        self.assertFalse(any(t.tool_name == "read_wiki_structure" for t in mcp.tools))
        mcp.disconnect()

    def test_tool_types(self):
        """
        Test that the tool types are correctly set when getting tools using .get_tools_as() method.
        Also test that mixing tool types in a Task raises IllogicalConfiguration.
        """
        from yacana.tool import IllogicalConfiguration
        mcp = Mcp("http://localhost:1337/mcp")
        mcp.connect()
        # Vérifie que tous les outils sont bien du type demandé
        for tool_type in [ToolType.OPENAI, ToolType.YACANA]:
            tools = mcp.get_tools_as(tool_type)
            for t in tools:
                self.assertEqual(t.tool_type, tool_type)
        # Teste qu'on ne peut pas mélanger les types d'outils dans une Task
        tools_openai = mcp.get_tools_as(ToolType.OPENAI)
        tools_yacana = mcp.get_tools_as(ToolType.YACANA)
        if tools_openai and tools_yacana:
            # Change dynamiquement le type d'un outil pour provoquer l'erreur
            mixed_tools = [tools_openai[0], tools_yacana[0]]
            # Force un type différent sur le deuxième outil
            mixed_tools[1].tool_type = ToolType.YACANA if mixed_tools[0].tool_type == ToolType.OPENAI else ToolType.OPENAI
            for _, agent in self._get_agents():
                with self.assertRaises(IllogicalConfiguration):
                    Task("Test mixing tool types", agent, tools=mixed_tools).solve()
        mcp.disconnect()

    def test_no_tools_are_used(self):
        """If the request doesn't match any given tools, the agent should not use a tool and answer directly."""
        mcp = Mcp("https://mcp.deepwiki.com/mcp")
        mcp.connect()
        tools = mcp.get_tools_as(ToolType.OPENAI)
        for _, agent in self._get_agents():
            msg = Task("What is the capital of France?", agent, tools=tools).solve()
            # On s'attend à une réponse directe, pas à un appel d'outil
            self.assertIn("Paris", msg.content)
        mcp.disconnect()

    def test_tools(self):
        """
        Calls the tools and make sure the output is as expected.
        Use tool execution types YACANA and OPENAI to test the tools.
        Use agent type Ollama and OpenAi to test the tools.
        """
        # On teste sur le serveur local pour éviter les quotas
        mcp = Mcp("http://localhost:1337/mcp")
        mcp.connect()
        # Test avec les deux types d'exécution
        for tool_type in [ToolType.OPENAI, ToolType.YACANA]:
            tools = mcp.get_tools_as(tool_type)
            for _, agent in self._get_agents():
                self._reset_agents()
                # Test de l'outil "weather"
                msg = Task("What is the weather in Paris?", agent, tools=tools).solve()
                self.assertTrue("Paris" in msg.content or "weather" in msg.content.lower())
                # Test de l'outil "example_tool"
                msg2 = Task("Call example_tool with name 'yacana'", agent, tools=tools).solve()
                self.assertIn("yacana", msg2.content.lower())
        mcp.disconnect()

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures before running tests."""
        super().setUpClass()
        # Get which agents to run from environment variables
        cls.run_ollama = os.getenv('TEST_OLLAMA', 'true').lower() == 'true'
        cls.run_openai = os.getenv('TEST_OPENAI', 'true').lower() == 'true'
        cls.run_vllm = os.getenv('TEST_VLLM', 'true').lower() == 'true'
        cls.run_lmstudio = os.getenv('TEST_LMSTUDIO', 'true').lower() == 'true'

if __name__ == '__main__':
    unittest.main()



import logging

from .generic_agent import GenericAgent


class Agent(GenericAgent):
    """
    Deprecated ! Do not use !
    Use OllamaAgent() or OpenAiAgent() instead depending on the backend you want to use.
    """

    def __init__(self, *args, **kwargs) -> None:
        deprecated_message = """
                Deprecation error: Use specialized Agents class instead. Sorry this had to be removed... But now that Yacana manages more than Ollama as backend, this class didn't make much sense.
                * If you are using Ollama as backend then the new `Agent` class name you should use is `OllamaAgent()`.
                * For OpenAi compatible backends (chatGPT, VLLM, etc), use `OpenAiAgent()`.
                """
        logging.fatal(deprecated_message)
        raise DeprecationWarning(deprecated_message)

    def _interact(self, task, tools, json_output, structured_output, medias, streaming_callback, task_runtime_config, tags):
        pass



import json
import logging
import uuid

from ollama import Client, ChatResponse
from typing import List, Type, Any, T, Dict, Callable, Tuple
from collections.abc import Iterator
from pydantic import BaseModel

from .generic_agent import GenericAgent
from .model_settings import OllamaModelSettings
from .utils import Dotdict, AgentType
from .exceptions import IllogicalConfiguration, TaskCompletionRefusal
from .history import HistorySlot, History
from .messages import GenericMessage, MessageRole, OllamaUserMessage, OllamaStructuredOutputMessage, OllamaTextMessage, ToolCallFromLLM, OpenAIFunctionCallingMessage
from .tool import Tool, ToolType
from .constants import PROMPT_TAG, RESPONSE_TAG

logger = logging.getLogger(__name__)


class OllamaAgent(GenericAgent):
    """
    Representation of an LLM agent that interacts with the Ollama inference server.

    This class provides ways to interact with the LLM, but it should not be controlled directly.
    Instead, it should be assigned to a Task(). When a task is required to be solved, the agent will
    interact with the prompt inside the task and output an answer. This class is more about
    configuring the agent than interacting with it.

    Parameters
    ----------
    name : str
        Name of the agent. Use something short and meaningful that doesn't contradict the system prompt.
    model_name : str
        Name of the LLM model that will be sent to the inference server (e.g., 'llama:3.1' or 'mistral:latest').
    system_prompt : str | None, optional
        Defines the way the LLM will behave (e.g., "You are a pirate" to have it talk like a pirate).
        Defaults to None.
    endpoint : str, optional
        The Ollama endpoint URL. Defaults to "http://127.0.0.1:11434".
    headers : dict, optional
        Custom headers to be sent with the inference request. Defaults to None.
    model_settings : OllamaModelSettings, optional
        All settings that Ollama currently supports as model configuration. Defaults to None.
    runtime_config : Dict | None, optional
        Runtime configuration for the agent. Defaults to None.
    thinking_tokens : Tuple[str, str] | None, optional
        A tuple containing the start and end tokens of a thinking LLM. For instance, "<think>" and "</think>" for Deepseek-R1.
        Setting this prevents the framework from getting sidetracked during the thinking steps and helps maintain focus on the final result.
    structured_thinking : bool, optional
        If True, Yacana will use structured_output internally to get better accuracy. If your LLM doesn't support structured_output set this to False.
        Defaults to True.
    **kwargs
        Additional keyword arguments passed to the parent class.

    Attributes
    ----------
    _agent_type : AgentType
        Type of the Agent to circumvent partial import when determining agent's type at runtime.

    Raises
    ------
    IllogicalConfiguration
        If model_settings is not an instance of OllamaModelSettings.
    """

    def __init__(self, name: str, model_name: str, system_prompt: str | None = None, endpoint: str = "http://127.0.0.1:11434", headers=None, model_settings: OllamaModelSettings = None, runtime_config: Dict | None = None, thinking_tokens: Tuple[str, str] | None = None, structured_thinking=True, **kwargs) -> None:
        model_settings = OllamaModelSettings() if model_settings is None else model_settings
        if not isinstance(model_settings, OllamaModelSettings):
            raise IllogicalConfiguration("model_settings must be an instance of OllamaModelSettings.")
        self._agent_type: AgentType = AgentType.OLLAMA
        super().__init__(name, model_name, model_settings, system_prompt=system_prompt, endpoint=endpoint, api_token="", headers=headers, runtime_config=runtime_config, history=kwargs.get("history", None), task_runtime_config=kwargs.get("task_runtime_config", None), thinking_tokens=thinking_tokens, structured_thinking=structured_thinking)

    def _interact(self, task: str, tools: List[Tool], json_output: bool, structured_output: Type[BaseModel] | None, medias: List[str] | None, streaming_callback: Callable | None = None, task_runtime_config: Dict | None = None, tags: List[str] | None = None) -> GenericMessage:
        """
        Main interaction method that handles task execution with optional tool usage.

        Parameters
        ----------
        task : str
            The task to execute.
        tools : List[Tool]
            List of available tools.
        json_output : bool
            Whether to output JSON.
        structured_output : Type[BaseModel] | None
            Optional structured output type.
        medias : List[str] | None
            Optional list of media files.
        streaming_callback : Callable | None, optional
            Optional callback for streaming responses. Defaults to None.
        task_runtime_config : Dict | None, optional
            Optional runtime configuration for the task. Defaults to None.
        tags : List[str] | None, optional
            Optional list of tags. Defaults to None.
        Returns
        -------
        GenericMessage
            The response message from the agent.

        Raises
        ------
        IllogicalConfiguration
            If streaming is requested with tool usage.
        """
        self._set_correct_tool_caller(tools)
        self._tags = tags if tags is not None else []
        tools: List[Tool] = [] if tools is None else tools
        self.task_runtime_config = task_runtime_config if task_runtime_config is not None else {}

        if len(tools) == 0:
            self._chat(self.history, task, medias=medias, json_output=json_output, structured_output=structured_output, streaming_callback=streaming_callback)
        elif len(tools) == 1:
            self.tool_caller.propose_tool(task, tools, json_output, structured_output, medias, streaming_callback, task_runtime_config, tags)
        elif len(tools) > 1:
            self.tool_caller.propose_tools(task, tools, json_output, structured_output, medias, streaming_callback, task_runtime_config, tags)

        return self.history.get_last_message()

    """def _ollama_tool_names_conversion(self, tools: List[Tool]) -> List[Tool]:
        for tool in tools:
            if not tool.is_mcp and tool.tool_type == ToolType.OPENAI:
                if tool.function_ref.__name__ != tool.tool_name:
                    logging.warning(f"Ollama expects the tool name to be the same as the function name. Tool '{tool.tool_name}' will be renamed automatically to '{tool.function_ref.__name__}'.")
                    tool.update_tool_name_to_match_function_name()

        return tools"""

    def _stream(self) -> None:
        """
        Placeholder for streaming functionality.

        This method is currently not implemented.
        """
        pass

    @staticmethod
    def _get_expected_output_format(json_output: bool, structured_output: Type[BaseModel] | None) -> dict[str, Any] | str:
        """
        Determines the expected output format (JSON, structured_output, etc.) based on the configuration set by the Task.

        Parameters
        ----------
        json_output : bool
            Whether to output JSON.
        structured_output : Type[BaseModel] | None
            Optional structured output type.

        Returns
        -------
        dict[str, Any] | str
            The expected output format.
        """
        if structured_output:
            return structured_output.model_json_schema()
        elif json_output:
            return 'json'
        else:
            return ''

    def _ollama_tool_call_to_json(self, tool_calls: List) -> List:
        tools_as_json: List = []
        if tool_calls:
            # There may be multiple tool calls in the response
            for tool in tool_calls:
                tools_as_json.append({"name": tool.function.name, "arguments": tool.function.arguments})
        return tools_as_json

    def _response_to_json(self, response: Any) -> str:
        """
        Converts an Ollama response object to JSON format. Used for setting raw_llm_json in HistorySlot.

        Parameters
        ----------
        response : Any
            The response object to convert.

        Returns
        -------
        str
            The JSON string representation of the response.

        Raises
        ------
        TypeError
            If the conversion to JSON fails.
        """
        try:
            result: Dict[str, Any] = {
                'model': getattr(response, 'model', None),
                'created_at': getattr(response, 'created_at', None),
                'done': getattr(response, 'done', None),
                'done_reason': getattr(response, 'done_reason', None),
                'total_duration': getattr(response, 'total_duration', None),
                'load_duration': getattr(response, 'load_duration', None),
                'prompt_eval_count': getattr(response, 'prompt_eval_count', None),
                'prompt_eval_duration': getattr(response, 'prompt_eval_duration', None),
                'eval_count': getattr(response, 'eval_count', None),
                'eval_duration': getattr(response, 'eval_duration', None),
            }

            # Extract 'message' if present
            message = getattr(response, 'message', None)
            if message is not None:
                result['message'] = {
                    'role': getattr(message, 'role', None),
                    'content': getattr(message, 'content', None),
                    'images': getattr(message, 'images', None),
                    'tool_calls': self._ollama_tool_call_to_json(getattr(message, 'tool_calls', None))
                }

            # Return the JSON string representation
            return json.dumps(result, indent=4)
        except Exception as e:
            raise TypeError(f"Failed to convert response to JSON: {e}")

    def _dispatch_chunk_if_streaming(self, chat_response: ChatResponse | Iterator[ChatResponse], streaming_callback: Callable | None) -> Dict | ChatResponse | Iterator[ChatResponse]:
        """
        Handles streaming responses by dispatching chunks to the callback.

        Parameters
        ----------
        chat_response : ChatResponse | Iterator[ChatResponse]
            The completion response or iterator.
        streaming_callback : Callable | None
            Optional callback for streaming responses.

        Returns
        -------
        Dict | ChatResponse | Iterator[ChatResponse]
            The processed response.

        Raises
        ------
        TaskCompletionRefusal
            If the streaming response contains no data.
        """
        # If we are not streaming, we return the Ollama message directly.
        if streaming_callback is None:
            return chat_response
        all_chunks = ""
        for chunk in chat_response:
            if chunk['message']['content'] is not None:
                all_chunks += chunk['message']['content']
                streaming_callback(chunk['message']['content'])
            else:
                raise TaskCompletionRefusal("Streaming LLMs response returned no data (content == None).")
        return Dotdict({
                    "message": {
                        "content": all_chunks,
                    }
                }
            )

    def _is_tool_calling(self, message) -> bool:
        """
        Checks if the message contains tool calls.

        Parameters
        ----------
        message : Any
            The message to check.

        Returns
        -------
        bool
            True if the choice contains tool calls, False otherwise.
        """
        return message.tool_calls and len(message.tool_calls) > 0

    def _chat(self, history: History, task: str | None, medias: List[str] | None = None, json_output: bool = False, structured_output: Type[T] | None = None, save_to_history: bool = True, tools: List[Tool] | None = None, streaming_callback: Callable | None = None) -> GenericMessage:
        """
        Main chat method that handles communication with the Ollama server.

        Parameters
        ----------
        history : History
            The conversation history.
        task : str | None
            The task to execute.
        medias : List[str] | None, optional
            Optional list of media files. Defaults to None.
        json_output : bool, optional
            Whether to output JSON. Defaults to False.
        structured_output : Type[T] | None, optional
            Optional structured output type. Defaults to None.
        save_to_history : bool, optional
            Whether to save the response to history. Defaults to True.
        tools : List[Tool] | None, optional
            Optional list of tools. Defaults to None.
        streaming_callback : Callable | None, optional
            Optional callback for streaming responses. Defaults to None.

        Returns
        -------
        GenericMessage
            The response content.
        """

        if task:
            logging.info(f"[PROMPT][To: {self.name}]: {task}")
            question_slot = history.add_message(OllamaUserMessage(MessageRole.USER, task, tags=self._tags + [PROMPT_TAG], medias=medias, structured_output=structured_output))

        if tools is not None and len(tools) > 0:
            for tool in tools:
                if tool.tool_type == ToolType.OPENAI and tool.optional is False and tool.shush is False:
                    logging.warning(f"You chose to use the OpenAI style tool calling with the OllamaAgent for the tool '{tool.tool_name}'. This tool is set by default as optional=False (hence making it mandatory to use). Note that Ollama does NOT support setting tools optional status on tools! They are all optional by default and this cannot be changed. Yacana may in the future mitigate this issue. If this is important for you please open an issue on the Yacana Github. You can hide this warning by setting `shush=True` in the Tool constructor.")
        client = Client(host=self.endpoint, headers=self.headers)
        params = {
            "model": self.model_name,
            "messages": history.get_messages_as_dict(),
            "format": self._get_expected_output_format(json_output, structured_output),
            "stream": True if streaming_callback is not None else False,
            "options": self.model_settings.get_settings(),
            **({"tools": [tool._openai_function_schema for tool in tools]} if tools is not None else {}),
            **self.runtime_config,
            **self.task_runtime_config
        }
        logging.debug("Runtime parameters before inference: %s", str(params))
        chat_response: ChatResponse | Iterator[ChatResponse] = client.chat(**params)
        response = self._dispatch_chunk_if_streaming(chat_response, streaming_callback)
        logging.debug("Inference output: %s", str(response))
        if structured_output is not None:
            logging.debug("Response assessment is structured output")
            answer_slot: HistorySlot = history.add_message(OllamaStructuredOutputMessage(MessageRole.ASSISTANT, str(response['message']['content']), structured_output.model_validate_json(response['message']['content']), tags=self._tags + [RESPONSE_TAG]))
        elif self._is_tool_calling(response['message']):
            logging.debug("Response assessment is tool calling")
            tool_calls: List[ToolCallFromLLM] = []
            for tool_call in response['message']["tool_calls"]:
                tool_calls.append(ToolCallFromLLM(str(uuid.uuid4()), tool_call.function.name, tool_call.function.arguments))
                logging.debug("Tool info : Name= %s, Arguments= %s", tool_call.function.name, tool_call.function.arguments)
            answer_slot: HistorySlot = history.add_message(OpenAIFunctionCallingMessage(tool_calls, tags=self._tags))
        else:
            answer_slot: HistorySlot = history.add_message(OllamaTextMessage(MessageRole.ASSISTANT, response['message']['content'], tags=self._tags + [RESPONSE_TAG]))

        self.task_runtime_config = {}
        answer_slot.set_raw_llm_json(self._response_to_json(response))

        logging.info(f"[AI_RESPONSE][From: {self.name}]: {answer_slot.get_message().get_as_pretty()}")

        last_message = history.get_last_message()
        if save_to_history is False:
            if task:
                history.delete_slot(question_slot)
            history.delete_slot(answer_slot)
        return last_message



import unittest
import os
from tests.test_base import BaseAgentTest
from yacana import Task
from pydantic import BaseModel

class CountryFact(BaseModel):
    """A fact about a country."""
    name: str
    fact: str

class Facts(BaseModel):
    """Collection of country facts."""
    countryFacts: list[CountryFact]

class TestStructuredOutput(BaseAgentTest):
    """Test structured output capabilities of all agent types."""

    def test_basic_structured_output(self):
        """Test basic structured output with a simple prompt."""
        prompt = "Tell me 3 facts about Canada."
        
        # Test Ollama agent
        if self.run_ollama:
            message = Task(prompt, self.ollama_agent, structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
        
        # Test OpenAI agent
        if self.run_openai:
            message = Task(prompt, self.openai_agent, structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
        
        # Test VLLM agent
        if self.run_vllm:
            message = Task(prompt, self.vllm_agent, structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)

    def test_structured_output_with_context(self):
        """Test structured output with additional context in the prompt."""
        prompt = """I'm writing a travel guide. Please provide 3 interesting facts about France.
        For each fact, include both the name of the fact and a detailed description.
        Make sure the facts are unique and interesting for tourists."""
        
        # Test Ollama agent
        if self.run_ollama:
            message = Task(prompt, self.ollama_agent, structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
            # Additional validation for context-specific requirements
            for fact in message.structured_output.countryFacts:
                self.assertGreater(len(fact.fact), 50, "Facts should be detailed for a travel guide")
        
        # Test OpenAI agent
        if self.run_openai:
            message = Task(prompt, self.openai_agent, structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
            for fact in message.structured_output.countryFacts:
                self.assertGreater(len(fact.fact), 50, "Facts should be detailed for a travel guide")
        
        # Test VLLM agent
        if self.run_vllm:
            message = Task(prompt, self.vllm_agent, structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
            for fact in message.structured_output.countryFacts:
                self.assertGreater(len(fact.fact), 50, "Facts should be detailed for a travel guide")

    def test_structured_output_with_media(self):
        """Test structured output with media input."""
        image_path = self.get_test_image_path("burger.jpg")
        prompt = "Tell me 3 facts about what is depicted in this image."
        
        # Test Ollama agent
        if self.run_ollama:
            message = Task(prompt, self.ollama_vision_agent, medias=[image_path], structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
            # Additional validation for media-specific requirements
            for fact in message.structured_output.countryFacts:
                # Check if the fact mentions food items from the image
                fact_text = f"{fact.name} {fact.fact}".lower()
                self.assertTrue(
                    any(keyword in fact_text for keyword in ['burger', 'hamburger', 'fries', 'french fries', 'food', 'meal', 'fast food']),
                    f"Fact should mention food items from the image, got: {fact_text}"
                )
        
        # Test OpenAI agent
        if self.run_openai:
            message = Task(prompt, self.openai_agent, medias=[image_path], structured_output=Facts).solve()
            self._validate_facts_response(message.structured_output)
            for fact in message.structured_output.countryFacts:
                fact_text = f"{fact.name} {fact.fact}".lower()
                self.assertTrue(
                    any(keyword in fact_text for keyword in ['burger', 'hamburger', 'fries', 'french fries', 'food', 'meal', 'fast food']),
                    f"Fact should mention food items from the image, got: {fact_text}"
                )
        
        # Test VLLM agent
        #if self.run_vllm:
        #    message = Task(prompt, self.vllm_agent, medias=[image_path], structured_output=Facts).solve()
        #    self._validate_facts_response(message.structured_output)
        #    for fact in message.structured_output.countryFacts:
        #        fact_text = f"{fact.name} {fact.fact}".lower()
        #        self.assertTrue(
        #            any(keyword in fact_text for keyword in ['burger', 'hamburger', 'fries', 'french fries', 'food', 'meal', 'fast food']),
        #            f"Fact should mention food items from the image, got: {fact_text}"
        #        )

    def _validate_facts_response(self, facts):
        """Validate the facts response structure."""
        self.assertIsInstance(facts, Facts)
        self.assertIsInstance(facts.countryFacts, list)
        self.assertEqual(len(facts.countryFacts), 3)
        for fact in facts.countryFacts:
            self.assertIsInstance(fact, CountryFact)
            self.assertIsInstance(fact.name, str)
            self.assertIsInstance(fact.fact, str)
            self.assertGreater(len(fact.name), 0)
            self.assertGreater(len(fact.fact), 0)

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures before running tests."""
        super().setUpClass()
        
        # Get which agents to run from environment variables
        cls.run_ollama = os.getenv('TEST_OLLAMA', 'true').lower() == 'true'
        cls.run_openai = os.getenv('TEST_OPENAI', 'true').lower() == 'true'
        cls.run_vllm = os.getenv('TEST_VLLM', 'true').lower() == 'true'

if __name__ == '__main__':
    unittest.main() 


from typing import List, Dict, Any
from abc import ABC


class ModelSettings(ABC):
    """
    Abstract base class for model settings.

    This class provides a base implementation for model settings with functionality
    for exporting settings, creating instances, and managing initial values.
    The constructor initializes the internal state for tracking initial values of settings. This means that you can update
    any setting and use .reset() to restore the initial values of all settings.
    """

    _registry = {}

    def __init__(self):
        self._initial_values = {}

    def __init_subclass__(cls, **kwargs):
        """
        Register a new subclass of ModelSettings.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments passed to the parent class.
        """
        super().__init_subclass__(**kwargs)
        ModelSettings._registry[cls.__name__] = cls

    def _export(self) -> Dict:
        """
        Export the current settings to a dictionary.

        Returns
        -------
        Dict
            A dictionary containing all current settings, excluding internal
            tracking values. The dictionary includes a 'type' field indicating
            the class name.
        """
        members = self.__dict__.copy()
        members["type"] = self.__class__.__name__
        members.pop("_initial_values", None)
        return members

    @staticmethod
    def create_instance(members: Dict):
        """
        Create a new instance of a ModelSettings subclass from a dictionary.

        Parameters
        ----------
        members : Dict
            Dictionary containing the settings and the 'type' field indicating
            which subclass to instantiate.

        Returns
        -------
        ModelSettings
            A new instance of the appropriate ModelSettings subclass.

        Raises
        ------
        KeyError
            If the 'type' field is missing or the class is not registered.
        """
        cls_name = members.pop("type")
        cls = ModelSettings._registry.get(cls_name)
        return cls(**members)

    def get_settings(self) -> dict:
        """
        Get all current settings as a dictionary.

        Returns
        -------
        dict
            A dictionary containing all non-None settings that don't start with '_'.
        """
        return {key: value for key, value in self.__dict__.items() if value is not None and not key.startswith("_")}

    def reset(self) -> None:
        """
        Reset all settings to their initial values.

        This method restores all settings to the values they had when the instance
        was first created.

        Notes
        -----
        This method iterates through all initial values stored during initialization
        and restores them to their original state.
        """
        for key, value in self._initial_values.items():
            setattr(self, f"{key}", value)


class OllamaModelSettings(ModelSettings):
    """
    Settings for Ollama model configuration.

    This class encapsulates all settings recognized by the Ollama inference server.
    Note that some settings may have no effect when using other inference servers.

    Parameters
    ----------
    mirostat : int, optional
        Controls the model's creativity level (0: off, 1: on, 2: extra on).
    mirostat_eta : float, optional
        Adjusts how quickly the model learns from context (e.g., 0.1).
    mirostat_tau : float, optional
        Controls topic adherence (e.g., 5.0).
    num_ctx : int, optional
        Determines context window size (e.g., 4096).
    num_gqa : int, optional
        Controls parallel task handling (e.g., 8).
    num_gpu : int, optional
        Sets GPU utilization (e.g., 50).
    num_thread : int, optional
        Controls parallel processing (e.g., 8).
    repeat_last_n : int, optional
        Controls repetition prevention window (e.g., 64).
    repeat_penalty : float, optional
        Penalty for repeated content (e.g., 1.1).
    temperature : float, optional
        Controls response randomness (e.g., 0.7).
    seed : int, optional
        Random seed for reproducibility (e.g., 42).
    stop : List[str], optional
        Stop sequences for generation.
    tfs_z : float, optional
        Controls response randomness reduction (e.g., 2.0).
    num_predict : int, optional
        Maximum tokens to generate (e.g., 128).
    top_k : int, optional
        Limits token selection (e.g., 40).
    top_p : float, optional
        Controls token selection probability (e.g., 0.9).
    **kwargs
        Additional settings passed to the parent class.

    Attributes
    ----------
    mirostat : int, optional
        Controls the model's creativity level (0: off, 1: on, 2: extra on).
    mirostat_eta : float, optional
        Adjusts how quickly the model learns from context (e.g., 0.1).
    mirostat_tau : float, optional
        Controls topic adherence (e.g., 5.0).
    num_ctx : int, optional
        Determines context window size (e.g., 4096).
    num_gqa : int, optional
        Controls parallel task handling (e.g., 8).
    num_gpu : int, optional
        Sets GPU utilization (e.g., 50).
    num_thread : int, optional
        Controls parallel processing (e.g., 8).
    repeat_last_n : int, optional
        Controls repetition prevention window (e.g., 64).
    repeat_penalty : float, optional
        Penalty for repeated content (e.g., 1.1).
    temperature : float, optional
        Controls response randomness (e.g., 0.7).
    seed : int, optional
        Random seed for reproducibility (e.g., 42).
    stop : List[str], optional
        Stop sequences for generation.
    tfs_z : float, optional
        Controls response randomness reduction (e.g., 2.0).
    num_predict : int, optional
        Maximum tokens to generate (e.g., 128).
    top_k : int, optional
        Limits token selection (e.g., 40).
    top_p : float, optional
        Controls token selection probability (e.g., 0.9).
    """

    def __init__(self,
                 mirostat: int = None,
                 mirostat_eta: float = None,
                 mirostat_tau: float = None,
                 num_ctx: int = None,
                 num_gqa: int = None,
                 num_gpu: int = None,
                 num_thread: int = None,
                 repeat_last_n: int = None,
                 repeat_penalty: float = None,
                 temperature: float = None,
                 seed: int = None,
                 stop: List[str] = None,
                 tfs_z: float = None,
                 num_predict: int = None,
                 top_k: int = None,
                 top_p: float = None,
                 **kwargs) -> None:
        super().__init__()
        # Initialize all properties
        self.mirostat = mirostat
        self.mirostat_eta = mirostat_eta
        self.mirostat_tau = mirostat_tau
        self.num_ctx = num_ctx
        self.num_gqa = num_gqa
        self.num_gpu = num_gpu
        self.num_thread = num_thread
        self.repeat_last_n = repeat_last_n
        self.repeat_penalty = repeat_penalty
        self.temperature = temperature
        self.seed = seed
        self.stop = stop
        self.tfs_z = tfs_z
        self.num_predict = num_predict
        self.top_k = top_k
        self.top_p = top_p

        # Store the initial values for resetting
        self._initial_values = {key: value for key, value in self.__dict__.items() if not key.startswith("_")}


class OpenAiModelSettings(ModelSettings):
    """
    Settings for OpenAI model configuration.

    This class encapsulates all settings recognized by the OpenAI API.
    Note that some settings may have no effect when using other inference servers.

    Parameters
    ----------
    audio : Any, optional
        Parameters for audio output when using audio modality.
    frequency_penalty : float, optional
        Penalty for token frequency (-2.0 to 2.0).
    logit_bias : Dict, optional
        Token bias adjustments (-100 to 100).
    logprobs : bool, optional
        Whether to return token log probabilities.
    max_completion_tokens : int, optional
        Maximum tokens to generate.
    metadata : Dict, optional
        Additional metadata (max 16 key-value pairs).
    modalities : List[str], optional
        Output types to generate (e.g., ["text", "audio"]).
    n : int, optional
        Number of completion choices to generate.
    prediction : Any, optional
        Configuration for predicted output.
    presence_penalty : float, optional
        Penalty for token presence (-2.0 to 2.0).
    reasoning_effort : str, optional
        Reasoning effort level ("low", "medium", "high").
    seed : int, optional
        Random seed for reproducibility.
    service_tier : str, optional
        Latency tier for processing ("auto" or "default").
    stop : str | List, optional
        Stop sequences for generation.
    store : bool, optional
        Whether to store completion output.
    stream_options : Any, optional
        Options for streaming response.
    temperature : float, optional
        Sampling temperature (0 to 2).
    top_logprobs : int, optional
        Number of top tokens to return (0 to 20).
    top_p : float, optional
        Nucleus sampling parameter.
    user : str, optional
        End-user identifier.
    web_search_options : Any, optional
        Web search configuration.
    **kwargs
        Additional settings passed to the parent class.

    Attributes
    ----------
    audio : Any, optional
        Parameters for audio output when using audio modality.
    frequency_penalty : float, optional
        Penalty for token frequency (-2.0 to 2.0).
    logit_bias : Dict, optional
        Token bias adjustments (-100 to 100).
    logprobs : bool, optional
        Whether to return token log probabilities.
    max_completion_tokens : int, optional
        Maximum tokens to generate.
    metadata : Dict, optional
        Additional metadata (max 16 key-value pairs).
    modalities : List[str], optional
        Output types to generate (e.g., ["text", "audio"]).
    n : int, optional
        Number of completion choices to generate.
    prediction : Any, optional
        Configuration for predicted output.
    presence_penalty : float, optional
        Penalty for token presence (-2.0 to 2.0).
    reasoning_effort : str, optional
        Reasoning effort level ("low", "medium", "high").
    seed : int, optional
        Random seed for reproducibility.
    service_tier : str, optional
        Latency tier for processing ("auto" or "default").
    stop : str | List, optional
        Stop sequences for generation.
    store : bool, optional
        Whether to store completion output.
    stream_options : Any, optional
        Options for streaming response.
    temperature : float, optional
        Sampling temperature (0 to 2).
    top_logprobs : int, optional
        Number of top tokens to return (0 to 20).
    top_p : float, optional
        Nucleus sampling parameter.
    user : str, optional
        End-user identifier.
    web_search_options : Any, optional
        Web search configuration.
    """

    def __init__(self,
                 audio: Any = None,
                 frequency_penalty: float = None,
                 logit_bias: Dict = None,
                 logprobs: bool = None,
                 max_completion_tokens: int = None,
                 metadata: Dict = None,
                 modalities: List[str] = None,
                 n: int = None,
                 prediction: Any = None,
                 presence_penalty: float = None,
                 reasoning_effort: str = None,
                 seed: int = None,
                 service_tier: str = None,
                 stop: str | List = None,
                 store: bool = None,
                 stream_options: Any = None,
                 temperature: float = None,
                 top_logprobs: int = None,
                 top_p: float = None,
                 user: str = None,
                 web_search_options: Any = None,
                 **kwargs) -> None:
        super().__init__()
        # Initialize all properties
        self.audio = audio
        self.frequency_penalty = frequency_penalty
        self.logit_bias = logit_bias
        self.logprobs = logprobs
        self.max_completion_tokens = max_completion_tokens
        self.metadata = metadata
        self.modalities = modalities
        self.n = n
        self.prediction = prediction
        self.presence_penalty = presence_penalty
        self.reasoning_effort = reasoning_effort
        self.seed = seed
        self.service_tier = service_tier
        self.stop = stop
        self.store = store
        self.stream_options = stream_options
        self.temperature = temperature
        self.top_logprobs = top_logprobs
        self.top_p = top_p
        self.user = user
        self.web_search_options = web_search_options


        # Store the initial values for resetting
        self._initial_values = {key: value for key, value in self.__dict__.items() if not key.startswith("_")}

    def reset(self) -> None:
        """
        Reset all settings to their initial values.

        This method restores all OpenAI model settings to the values they had when
        the instance was first created.

        Notes
        -----
        This method iterates through all initial values stored during initialization
        and restores them to their original state. This includes all OpenAI-specific
        parameters like temperature, top_p, and other generation settings.
        """
        for key, value in self._initial_values.items():
            setattr(self, f"{key}", value)




import logging
from typing import List, TypedDict
import regex

from .exceptions import SpecializedTokenCountingError


class HuggingFaceDetails:
    def __init__(self, repo_name: str, token: str | None = None):
        self.repo_name = repo_name
        self.token = token


class HFMessage(TypedDict):
    role: str
    content: str


def count_tokens_using_huggingface(hugging_faces_messages: List[HFMessage], hugging_face_repo_name: str, hugging_face_token: str | None = None) -> int:
    try:
        from transformers import AutoTokenizer
        if hugging_face_token:
            from huggingface_hub import login
            logging.debug("Logging into Hugging Face Hub to access private model for token counting. This may take some time...")
            login(hugging_face_token)
        logging.debug("Loading tokenizer from Hugging Face Hub for model: " + hugging_face_repo_name)
        tokenizer = AutoTokenizer.from_pretrained(hugging_face_repo_name)
        tokens = tokenizer.apply_chat_template(hugging_faces_messages, tokenize=True)
        nb_tokens = len(tokens)
        return nb_tokens
    except Exception as e:
        logging.warning(f"Could not load tokenizer for model {hugging_face_repo_name}. Falling back to approximative token count. Error: {e}")
        raise SpecializedTokenCountingError(str(e))


def count_tokens_using_tiktoken(llm_model_name: str, message: HFMessage) -> int:
    import tiktoken
    try:
        logging.debug("Loading tiktoken encoding for model: " + llm_model_name)
        enc = tiktoken.encoding_for_model(llm_model_name)  # Getting correct encoding if it's an OpenAI model ONLY else ValueError
        return len(enc.encode(message["role"])) + len(enc.encode(message["content"]))
    except KeyError:
        logging.debug(f"Could not find encoding for model {llm_model_name}. This is normal if this model is not from OpenAI. You should set the @hugging_face_repo_nama and token in the Agent class so Yacana may use the correct tokeniser for this LLM. Falling back to approximative token count instead.")
        raise SpecializedTokenCountingError(f"Could not find encoding for model {llm_model_name}.")


def count_tokens_using_regex(message: HFMessage) -> int:
    token_match_regex = regex.compile(r"(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+", regex.MULTILINE)
    logging.debug("Using approximative token count method.")
    return len(token_match_regex.findall(message["role"]) + token_match_regex.findall(message["content"]))



import copy
from enum import Enum, unique
from typing import List
import logging
from typing_extensions import Tuple

from .messages import GenericMessage, MessageRole
from .history import Message
from .task import Task
from .exceptions import ReachedTaskCompletion, IllogicalConfiguration

logger = logging.getLogger(__name__)


@unique
class EndChatMode(Enum):
    """
    <b>ENUM:</b> All types of group chat completion.

    The difficulty in making agents talk to each other is not to have them talk but to have them stop talking.
    Note that only tasks that have the @llm_stops_by_itself=True are actually impacted by the mode set here.
    Use in conjunction with EndChat().

    Attributes
    ----------
    ALL_TASK_MUST_COMPLETE : str
        Chat will continue going until all LLMs with @llm_stops_by_itself=True says they are finished.
        Set precise completion goals in the task prompt if you want this to actually work.
    ONE_LAST_CHAT_AFTER_FIRST_COMPLETION : str
        One agent will have the opportunity to respond after the completion of one agent allowing it to answer one last time.
    ONE_LAST_GROUP_CHAT_AFTER_FIRST_COMPLETION : str
        All agents will have one last table turn to speak before exiting the chat after the first completion arrives.
    END_CHAT_AFTER_FIRST_COMPLETION : str
        Immediately stops group chat after an agent has reached completion.
    MAX_ITERATIONS_ONLY : str
        Agents won't be asked if they have fulfilled their objectives but instead will loop until achieving max iteration.
        Max iteration can be set in the EndChat() class.
    """
    ALL_TASK_MUST_COMPLETE = "ALL_TASK_MUST_COMPLETE"
    ONE_LAST_CHAT_AFTER_FIRST_COMPLETION = "ONE_LAST_CHAT_AFTER_FIRST_COMPLETION"
    ONE_LAST_GROUP_CHAT_AFTER_FIRST_COMPLETION = "ONE_LAST_GROUP_CHAT_AFTER_FIRST_COMPLETION"
    END_CHAT_AFTER_FIRST_COMPLETION = "END_CHAT_AFTER_FIRST_COMPLETION"
    MAX_ITERATIONS_ONLY = "MAX_ITERATIONS_ONLY"


class EndChat:
    """
    Defines the modality of how and when LLMs stop chatting.

    By default, when reaching the @max_iterations limit the chat will end.
    However, if a task has set @llm_stops_by_itself=True then you can use one of the EndChatMode values to specify how
    and when the chat stops after reaching the first completion.

    Parameters
    ----------
    mode : EndChatMode
        The modality to end a chat with multiple agents.
    max_iterations : int, optional
        The max number of iterations in a conversation. An iteration is complete when we get back to the first speaker.
        Defaults to 5.

    Raises
    ------
    ValueError
        If mode is not an instance of EndChatMode.
        If max_iterations is less than or equal to 0.
    """
    def __init__(self, mode: EndChatMode, max_iterations: int = 5):
        if not isinstance(mode, EndChatMode):
            raise ValueError(
                "@mode must be an instance of EndChatMode. For instance: EndChatMode.ALL_TASK_MUST_COMPLETE")
        if max_iterations <= 0:
            raise ValueError("@max_iterations must be > 0")
        self.mode: EndChatMode = mode
        self.max_iterations: int = max_iterations


class GroupSolve:
    """
    This class allows multiple agents to enter a conversation with each other.

    This is made different from other frameworks as it is not directly the agents that enter the group chat but the tasks.
    Note that each task has an LLM Agent bound to it though. However, this changes the approach you should have when
    creating the conversation prompts.

    If you are letting LLMs decide when to stop chatting then be sure to specify precise goals. For instance
    'Your task is done when XXXX'. This way the LLM will reflect on that and make a decision if it has achieved
    its original Task.

    Note that dual chat is more efficient as each agent thinks it is talking to a human and this is how instruct models
    have been trained. When having a conversation with more than two agents (aka more than 2 tasks) then Yacana will
    force them into roleplay. This might degrade performance a bit but has the advantage of showing a clearer logs.
    Logging output for dual chat will be reworked shortly.

    Parameters
    ----------
    tasks : List[Task]
        All tasks that must be solved during group chat.
    end_chat : EndChat
        Defines the modality of how and when LLMs stop chatting.
    reconcile_first_message : bool, optional
        Should the first message from both LLMs be available to one another. Only useful in dual chat.
        Defaults to False.
    shift_message_owner : Task, optional
        The Task to which the shift message should be assigned to. In the end it's rather the corresponding Agent
        than the Task that is involved here. Defaults to None.
    shift_message_content : str, optional
        A custom message instead of using the opposite agent response as shift message content. Defaults to None.

    Attributes
    ----------
    tasks : List[Task]
        All tasks that must be solved during group chat.
    mode : EndChatMode
        The modality to end a chat with multiple agents.
    reconcile_first_message : bool
        Should the first message from both LLMs be available to one another. Only useful in dual chat.
    max_iter : int
        The max number of iterations in a conversation. An iteration is complete when we get back to the first speaker.
    shift_owner : Task
        The Task to which the shift message should be assigned to. In the end it's rather the corresponding Agent
    shift_content : str | None
        A custom message instead of using the opposite agent response as shift message content.

    Raises
    ------
    ValueError
        If tasks is None.
        If end_chat is not an instance of EndChat.
    """

    def __init__(self, tasks: List[Task], end_chat: EndChat, reconcile_first_message: bool = False, shift_message_owner: Task = None, shift_message_content: str | None = None) -> None:
        if tasks is None:
            raise ValueError("@tasks cannot be None. Must be a List of Task")
        if not isinstance(end_chat, EndChat):
            raise ValueError(
                "@end_chat_mode must be an instance of EndChat.")
        self.tasks: List[Task] = tasks
        self.mode: EndChatMode = end_chat.mode
        self.reconcile_first_message: bool = reconcile_first_message
        self.max_iter: int = end_chat.max_iterations
        self._once: bool = False
        self.shift_owner: Task = shift_message_owner
        self.shift_content: str | None = shift_message_content
        # Looks like this {"task_uid1": False, "task_uid2": False}
        self._tasks_completion_statuses = {cur_task.uuid: False for cur_task in self.tasks if
                                           cur_task.llm_stops_by_itself}

    def solve(self) -> None:
        """
        Starts the group chat and allows all LLMs to solve their assigned tasks.

        Note that 'dual chat' and '3 and more' chat have a different way of starting.
        Refer to the official documentation for more details.

        Raises
        ------
        IllogicalConfiguration
            If trying to use the same GroupSolve instance twice.
            If using an endChatMode that requires llm_stops_by_itself=True but no tasks have it set.
        IndexError
            If less than 2 tasks are provided.
        """
        if self._once is True:
            raise IllogicalConfiguration("You cannot use the same GroupSolve twice. You must instantiate a new one.")
        self._once = True
        # When @exit_after_next_chat passes True we'll exit after the next chat with an agent
        exit_after_next_chat: bool = False
        # When @exit_after_next_chat changes from None to Int then we'll exit when it reaches 0 as we would
        # have completed one additional full chat iteration
        exit_after_next_group_chat: int | None = None

        if (self.mode != EndChatMode.MAX_ITERATIONS_ONLY) and any(
                [cur_task.llm_stops_by_itself for cur_task in self.tasks]) is False:
            raise IllogicalConfiguration(
                "You set an endChatMode in the GroupSolve() class that requires at least one Task() having the @llm_stops_by_itself argument set to True in its constructor. None of your tasks have it set to True.")

        if len(self.tasks) <= 1:
            raise IndexError("Must give at least 2 Tasks to make a group.")

        elif len(self.tasks) == 2:
            my_task1 = self.tasks[0]
            my_task2 = self.tasks[1]
            try:
                # First speaker
                my_task1.solve()

                # Giving prompt and AI output of first speaker to the second speaker
                if self.reconcile_first_message is True:
                    my_task2.agent.history.add_message(Message(MessageRole.USER, my_task1.prompt))
                    my_task2.agent.history.add_message(Message(MessageRole.ASSISTANT, my_task1.agent.history.get_last_message().content))
                # Second speaker (has the history of the first speaker)
                my_task2.solve()

                if self.reconcile_first_message is True:
                    my_task1.agent.history.add_message(Message(MessageRole.USER, my_task2.prompt))
                    my_task1.agent.history.add_message(Message(MessageRole.ASSISTANT, my_task2.agent.history.get_last_message().content))

                # (str, (TaskX, TaskY))
                last_generated_answer, tasks_run_order = self._set_shift_message()

                while self.max_iter > 0 and (exit_after_next_group_chat is None or exit_after_next_group_chat > 0):
                    logging.debug("Entering groupSolve.")

                    for current_task in tasks_run_order:
                        copy_task: Task = self._duplicate_task(current_task, last_generated_answer)
                        last_generated_answer: str = self._solve_copy(copy_task)
                        exit_after_next_chat, exit_after_next_group_chat = self._check_if_must_exit(exit_after_next_chat,
                                                                                                    exit_after_next_group_chat,
                                                                                                    copy_task)
                    logging.debug("One full iteration has been done (the two agents spoke)")
                    self.max_iter -= 1
                    logging.debug(f"Maximum iterations is now at {str(self.max_iter)}")
                    if isinstance(exit_after_next_group_chat, int):
                        exit_after_next_group_chat -= 1
            except ReachedTaskCompletion:
                pass

        elif len(self.tasks) >= 3:
            self._add_roleplay_prompts()

            for cur_task in self.tasks:
                # Changing the way the task is written to match the multi-agent format
                user_message: GenericMessage = Message(MessageRole.USER,
                                                f"[TaskManager]: {cur_task.agent.name}: this is your main task: `" + cur_task.prompt + "`")
                copy_task: Task = self._duplicate_task(cur_task, user_message.content)
                last_generated_answer: str = self._solve_copy(copy_task)
                self._reconcile_history(cur_task, user_message, last_generated_answer)
                exit_after_next_chat, exit_after_next_group_chat = self._check_if_must_exit(exit_after_next_chat,
                                                                                            exit_after_next_group_chat,
                                                                                            copy_task)

            try:
                while self.max_iter > 0 and (exit_after_next_group_chat is None or exit_after_next_group_chat > 0):
                    for cur_task in self.tasks:
                        user_message: GenericMessage = Message(MessageRole.USER,
                                                        f"[TaskManager]: {cur_task.agent.name} it's your turn to speak now.")
                        copy_task: Task = self._duplicate_task(cur_task, user_message.content)
                        last_generated_answer: str = self._solve_copy(copy_task)
                        self._reconcile_history(cur_task, user_message, last_generated_answer)
                        exit_after_next_chat, exit_after_next_group_chat = self._check_if_must_exit(
                            exit_after_next_chat, exit_after_next_group_chat, copy_task)
                    self.max_iter -= 1
                    if isinstance(exit_after_next_group_chat, int):
                        exit_after_next_group_chat -= 1
            except ReachedTaskCompletion:
                pass

    def _set_shift_message(self) -> tuple:
        """
        Assigns shift message to either Agents and sets a custom message if user provided one.
        If no custom message is provided, the opposite agent's ANSWER will be used as PROMPT (shift message).

        Returns
        -------
        tuple
            A tuple containing:
            - str: The last generated answer
            - tuple[Task, Task]: The tasks run order

        Raises
        ------
        IllogicalConfiguration
            If shift_message_owner is not a Task present in the tasks array.
        """
        my_task1 = self.tasks[0]
        my_task2 = self.tasks[1]

        # Setting default: shift message will be assigned to task2
        if self.shift_owner is None:
            self.shift_owner = my_task2

        # Assigning shift message to either Agents and sets a custom message if user gave one. If not, the opposite agent's ANSWER will be used as PROMPT (shift message).
        if self.shift_owner is my_task1:
            if self.shift_content is None:
                last_generated_answer: str = my_task2.agent.history.get_last_message().content
            else:
                last_generated_answer: str = self.shift_content
            tasks_run_order: Tuple = (my_task1, my_task2)
        elif self.shift_owner is my_task2:
            if self.shift_content is None:
                last_generated_answer: str = my_task1.agent.history.get_last_message().content
            else:
                last_generated_answer: str = self.shift_content
            tasks_run_order: Tuple = (my_task2, my_task1)
        else:
            raise IllogicalConfiguration("@shift_message_owner parameter from GroupSolve() class must be an instance of Task present in the array of Task also given to the constructor.")
        return last_generated_answer, tasks_run_order

    def _duplicate_task(self, task_to_duplicate: Task, last_generated_answer: str) -> Task:
        """
        Creates a new instance of an existing task.
        It will have the exact same configuration. However, we change the task prompt so it can solve something new.

        Parameters
        ----------
        task_to_duplicate : Task
            The Task to duplicate.
        last_generated_answer : str
            The prompt to change in the new Task.

        Returns
        -------
        Task
            The copy of the Task with the new prompt.
        """
        copy_task2: Task = copy.deepcopy(task_to_duplicate)
        copy_task2.prompt = last_generated_answer
        copy_task2.agent = task_to_duplicate.agent
        return copy_task2

    def _solve_copy(self, copy_task: Task) -> str:
        """
        Solves a copy of a task and returns its content.

        Parameters
        ----------
        copy_task : Task
            The task copy to solve.

        Returns
        -------
        str
            The content of the solved task's response.
        """
        last_generated_answer = copy_task.solve().content
        return last_generated_answer

    def _check_if_must_exit(self, exit_after_next_chat: bool, exit_after_next_group_chat: int | None, copy_task: Task) -> tuple[bool, int | None]:
        """
        Checks if the conversation should end based on various conditions.

        Parameters
        ----------
        exit_after_next_chat : bool
            Whether to exit after the next chat.
        exit_after_next_group_chat : int | None
            Number of group chats remaining before exit, or None if not applicable.
        copy_task : Task
            The current task being processed.

        Returns
        -------
        tuple[bool, int | None]
            Updated exit conditions after checking current state.

        Raises
        ------
        ReachedTaskCompletion
            If the task completion conditions are met.
        """
        if exit_after_next_chat is True:
            raise ReachedTaskCompletion()

        if copy_task.llm_stops_by_itself is True:
            if copy_task.use_self_reflection is False:
                history_save = copy.deepcopy(copy_task.agent.history)
            # @copy_task is already a copy and won't be reused so no use to deepcopy it again.
            copy_task.prompt = ("[TaskManager]: " if len(self.tasks) >= 3 else "") + "In your opinion, what objectives from your initial task have you NOT completed ?"

            Task(copy_task.prompt, copy_task.agent).solve()
            if copy_task.use_self_reflection is True:
                history_save = copy.deepcopy(copy_task.agent.history)

            ms_temp_save: float = copy_task.agent.model_settings.temperature
            copy_task.agent.model_settings.temperature = 0.3
            answer: str = Task("To summarize in one word, did you still had some objectives to fulfill ? Answer ONLY by 'yes' or 'no'. Do not output anything else !", copy_task.agent, forget=True).solve().content

            copy_task.agent.model_settings.temperature = ms_temp_save
            copy_task.agent.history = history_save
            if "yes" in answer.lower():
                logging.debug(f"Still got objectives to fulfill ? SAID YES: {answer}")
            elif "no" in answer.lower():
                logging.debug(f"Still got objectives to fulfill ? SAID NO: {answer}")
                if self.mode == EndChatMode.END_CHAT_AFTER_FIRST_COMPLETION:
                    raise ReachedTaskCompletion()
                if self.mode == EndChatMode.ONE_LAST_CHAT_AFTER_FIRST_COMPLETION:
                    exit_after_next_chat = True
                if self.mode == EndChatMode.ONE_LAST_GROUP_CHAT_AFTER_FIRST_COMPLETION:
                    if exit_after_next_group_chat is None:
                        # This goes like 'None -> 2 -> 1 -> 0 -> if 0 then exit'. But I don't like the fact that the variable carries two pieces of information at a time. Checking if it's a None or an int and basing business logic on the result isn't great.
                        exit_after_next_group_chat = len(self.tasks)
                if self.mode == EndChatMode.ALL_TASK_MUST_COMPLETE:
                    self._tasks_completion_statuses[copy_task.uuid] = True
                    if all(self._tasks_completion_statuses.values()) is True:
                        raise ReachedTaskCompletion()
            else:
                logging.error("While checking if the task was completed Yacana couldn't determine if the LLM said Yes or No. As a decision must be made Yacana will assume it is a Yes. If you encounter this frequently please open an issue on the github. Thx !")
        return exit_after_next_chat, exit_after_next_group_chat

    def _reconcile_history(self, cur_task: Task, user_message: GenericMessage, last_generated_answer: str) -> None:
        """
        Adds task and answer to all agents except the one who just spoke to reconcile histories.

        Parameters
        ----------
        cur_task : Task
            The current task that just spoke.
        user_message : GenericMessage
            The user message to add to other agents' histories.
        last_generated_answer : str
            The last generated answer to add to other agents' histories.
        """
        # Adding task + answer to everyone except the one who just spoke in order to reconcile histories with all other agents
        for cur_task2 in self.tasks:
            if cur_task2.uuid != cur_task.uuid:
                cur_task2.agent.history.add_message(user_message)
                cur_task2.agent.history.add_message(Message(MessageRole.ASSISTANT, last_generated_answer))

    def _add_roleplay_prompts(self) -> None:
        """
        Adds roleplay prompts to all tasks in the group.

        This method sets up the roleplay context for each agent, including their speaker name
        and the names of other participants in the conversation.
        """
        for cur_task in self.tasks:
            other_speaker_names = ",".join(list(filter(None, [
                f"[{cur_task2.agent.name}]" if cur_task2.uuid != cur_task.uuid else "" for cur_task2 in self.tasks])))
            cur_task.agent.history.add_message(Message(MessageRole.USER,
                                                     f"[TaskManager]: You are entering a roleplay with multiple "
                                                     f"speakers where each one has his own objectives to fulfill. "
                                                     f"Each message must follow this syntax '[speaker_name]: "
                                                     f"message'.\n" + f"The other speakers are: "
                                                                      f"{other_speaker_names}.\nYour "
                                                                      f"speaker name is [{cur_task.agent.name}].\nI "
                                                                      f"will give you your task in the next message."))
            cur_task.agent.history.add_message(Message(MessageRole.ASSISTANT,
                                                     f"[{cur_task.agent.name}]: Received and acknowledged. I'm ready "
                                                     f"to execute my tasks as {cur_task.agent.name}. Please proceed "
                                                     f"with my assignment."))




import inspect
import json
import logging
from enum import Enum, unique
from typing import List, Callable, Dict, Any, Tuple

from .exceptions import IllogicalConfiguration, McpBadToolConfig
from .function_to_json_schema import function_to_json_with_pydantic
from .history import History
from .messages import MessageRole, OllamaUserMessage


@unique
class ToolType(Enum):
    """
    <b>ENUM:</b> ToolType

    How a tool must be presented to the LLM so it can be called.

    Attributes
    ----------
    OPENAI : str
        Tool calling will follow the OpenAi function calling format.
    YACANA : str
        Tool calling will follow the Yacana function calling format available to all LLMs.
    """
    OPENAI = "OPENAI"
    YACANA = "YACANA"


class Tool:
    """
    A class representing a tool that can be used by an LLM to perform specific tasks.

    A tool is a classic python function that can be assigned to a task. The LLM in charge
    of solving the task will have access to the tool and may try to use it or not
    depending on the configuration provided.

    Parameters
    ----------
    tool_name : str
        A name for the tool. Should be concise and related to what the tool does.
    function_description : str
        A description for the tool. Should be concise and related to what the tool does.
        May contain an example of how to use. Refer to the documentation.
    function_ref : Callable
        The reference to a python function that will be called with parameters provided by the LLM.
    optional : bool, optional
        Allows to a certain extent the LLM to choose to use the given tool or not depending on the task to solve.
        Defaults to False.
    usage_examples : List[dict], optional
        A list of python dictionary examples of how the tool should be called.
        The examples will be given to the LLM to help it call the tool correctly.
        Use if the LLM struggles to call the tool successfully. Defaults to an empty list.
    max_custom_error : int, optional
        The max errors a tool can raise.
        A tool should raise a ToolError(...) exception with a detailed explanation of why it failed.
        The LLM will get the exception message and try again, taking into account the new knowledge it gained from the error.
        When reaching the max iteration the MaxToolErrorIter() exception is thrown and the task is stopped. Defaults to 5.
    max_call_error : int, optional
        The max number of times Yacana can fail to call a tool correctly.
        Note that Yacana uses the parameters given to the LLM to call the tool so if they are invalid then Yacana will have a hard time to fix the situation.
        You should try to give examples to the LLM on how to call the tool either in the tool description or when using the @usage_examples attribute to help the model.
        Defaults to 5.
    shush : bool, optional
        If True, the tool won't warn anymore about the unsupported tool setting optional=True and using Ollama + OpenAi style tool calling.
        Defaults to False.

    Attributes
    ----------
    tool_name : str
        The name of the tool.
    function_description : str
        A description of the tool's functionality.
    function_ref : Callable
        Function reference that the tool will call.
    optional : bool
        Indicates if the tool is optional.
    usage_examples : List[dict]
        A list of usage examples for the tool. The dict keys should match the function parameters.
    max_custom_error : int
        Maximum number of custom errors (raised from the function) allowed before stopping the task.
    max_call_error : int
        Maximum number of call errors (eg: python can't find the function) allowed before stopping the task.
    is_mcp: bool
        Is this tool a local tool or an MCP tool.
    shush : bool
        If True, suppresses warnings about unsupported optional tool configurations with Ollama.
    tool_type: ToolType
        The tool execution style. Either use the Yacana style (default) or follow the OpenAi style.

    Raises
    ------
    IllogicalConfiguration
        If max_custom_error or max_call_error is less than 0.
    """

    def __init__(self, tool_name: str, function_description: str, function_ref: Callable, optional: bool = False,
                 usage_examples: List[dict] | None = None, max_custom_error: int = 5, max_call_error: int = 5,
                 tool_type: ToolType = ToolType.YACANA, mcp_input_schema: dict = None, shush=False) -> None:
        self.tool_name: str = tool_name
        self.function_description: str = function_description
        self.function_ref: Callable = function_ref
        self.is_mcp: bool = mcp_input_schema is not None
        self.optional: bool = optional
        self.usage_examples: List[dict] = usage_examples if usage_examples is not None else []
        self.mcp_input_schema: dict = mcp_input_schema

        if mcp_input_schema is not None:
            self._openai_function_schema: dict = self._function_to_json_with_mcp(mcp_input_schema)
            params: List[Tuple[str, str]] = self.input_shema_to_prototype(mcp_input_schema)
            self._function_prototype: str = self.tool_name + "(" + ", ".join([f"{name}: {type_}" for name, type_ in params]) + ")"
            self._function_args: List[str] = [param[0] for param in params]
        else:
            self._openai_function_schema: dict = self._function_to_json_with_pydantic()
            self._function_prototype: str = Tool._extract_prototype(function_ref)
            self._function_args: List[str] = Tool._extract_parameters(function_ref)

        self.max_custom_error: int = max_custom_error
        self.max_call_error: int = max_call_error
        self.tool_type: ToolType = tool_type
        self.shush = shush

        if max_custom_error < 0 or max_call_error < 0:
            raise IllogicalConfiguration("@max_custom_error and @max_call_error must be > 0")
        if " " in self.tool_name:
            logging.warning(f"Tool name {self.tool_name} contains spaces. Some inference servers may not support it. We recommend you use CamelCase instead.")

    def input_shema_to_prototype(self, input_shema: dict) -> List[Tuple[str, str]]:
        """
        Converts the input schema to a function prototype string.

        Parameters
        ----------
        input_shema : dict
            The input schema to convert.

        Returns
        -------
         List[(str, str)]
            tuple[0] is the param name and tuple[1] is the param type.
        """
        if input_shema.get("type") != "object" or "properties" not in input_shema:
            raise McpBadToolConfig(f"For tool '{self.tool_name}' from source MCP : Input schema must be an object with properties.")

        result: List[Tuple[str, str]] = []
        for param_name, param_info in input_shema["properties"].items():
            param_type = param_info.get("type", "Any")
            result.append((param_name, param_type))

        return result

    def _get_examples_as_history(self, tags: List[str]) -> History:
        """
        Convert the tool's usage examples into a conversation history format.
        This is multi shot prompting based on examples given to the tool.

        Parameters
        ----------
        tags : List[str]
            A list of tags to add to the messages.

        Returns
        -------
        History
            A history object containing the tool usage examples formatted as a conversation.
        """
        history = History()

        for example in self.usage_examples:
            tmp = ", ".join([f"{key} is {value}" for key, value in example.items()])
            history.add_message(OllamaUserMessage(MessageRole.USER,
                                f"For training purpose let's try calling the tool {self.tool_name} with theses parameter{'s' if len(example.items()) > 1 else ''}: {tmp}", tags=tags))
            history.add_message(OllamaUserMessage(MessageRole.ASSISTANT, json.dumps(example), tags=tags))
        if len(self.usage_examples) > 0:
            history.add_message(OllamaUserMessage(MessageRole.USER,
                                f"{'These were all' if len(self.usage_examples) > 1 else 'This was a'} great tool call{'s' if len(self.usage_examples) > 1 else ''}", tags=tags))
            history.add_message(OllamaUserMessage(MessageRole.ASSISTANT, "Great ! I understand how it works.", tags=tags))
        return history

    @staticmethod
    def _extract_prototype(func: Callable) -> str:
        """
        Extract the function prototype as a string.

        Parameters
        ----------
        func : Callable
            The function to extract the prototype from.

        Returns
        -------
        str
            The function prototype as a string, including the function name and signature.
        """
        # Get the function's signature
        sig = inspect.signature(func)
        # Format the signature as a string and returns it
        return f"{func.__name__}{sig}"

    @staticmethod
    def _extract_parameters(func: Callable) -> List[str]:
        """
        Extract the parameter names from a function's signature.

        Parameters
        ----------
        func : Callable
            The function to extract parameters from.

        Returns
        -------
        List[str]
            A list of parameter names from the function's signature.
        """
        signature = inspect.signature(func)
        # Access the parameters
        parameters = signature.parameters
        # Extract the parameter names into a list
        return [param_name for param_name in parameters]

    def _function_to_json_with_pydantic(self) -> Dict:
        """
        Convert the function to a JSON schema using Pydantic.

        This method generates a JSON schema for the function that can be used by
        OpenAI's function calling API. The schema is stored in the
        _openai_function_schema attribute.
        """
        return function_to_json_with_pydantic(self.tool_name, self.function_description, self.function_ref)

    def _function_to_json_with_mcp(self, input_shema: dict) -> Dict:
        return {
            "type": "function",
            "function": {
                "name": self.tool_name,
                "description": self.function_description,
                "parameters": input_shema,
                "strict": True
            }
        }

    @staticmethod
    def bulk_tool_type_update(tools: List['Tool'], tool_type: ToolType) -> None:
        """
        Update the tool type for a list of tools.
        !Warning!: the tool type will remain the same until it is changed otherwise.
        This is different from mcp.get_tools_as(<tool_type>) which returns a copy of the tools.

        Parameters
        ----------
        tools : List[Tool]
            The list of tools to update.
        tool_type : ToolType
            The new tool type to set for all tools in the list.
        """
        for tool in tools:
            tool.tool_type = tool_type



from pydantic import BaseModel


class UseOtherTool(BaseModel):
    shouldUseOtherTool: bool


class UseTool(BaseModel):
    useTool: bool


class MakeAnotherToolCall(BaseModel):
    makeAnotherToolCall: bool



from abc import abstractmethod, ABC
from typing import List, Type, Callable, Dict
from pydantic import BaseModel

from .tool import Tool


class BaseToolCaller(ABC):
    def __init__(self, agent):
        self.agent = agent  # full access to _chat(), config, tools, etc.

    @abstractmethod
    def propose_tool(self, task: str, tools: List[Tool], json_output: bool, structured_output: Type[BaseModel] | None, medias: List[str] | None, streaming_callback: Callable | None = None, task_runtime_config: Dict | None = None, tags: List[str] | None = None):
        pass

    @abstractmethod
    def propose_tools(self, task: str, tools: List[Tool], json_output: bool, structured_output: Type[BaseModel] | None, medias: List[str] | None, streaming_callback: Callable | None = None, task_runtime_config: Dict | None = None, tags: List[str] | None = None):
        pass



class ToolError(Exception):
    """
    Exception raised when a tool encounters incorrect parameters.

    This exception is raised by the user from inside a tool when the tool determines
    that the given parameters are incorrect. The error message should be didactic
    to help the LLM fix its mistakes, otherwise it may loop until reaching
    MaxToolErrorIter.

    Parameters
    ----------
    message : str
        A descriptive message that will be given to the LLM to help fix the issue.
        Example: 'Parameter xxxx expected type int but got type string'.

    Notes
    -----
    The error message should be clear and instructive to help the LLM understand
    and correct the parameter issue.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class MaxToolErrorIter(Exception):
    """
    Exception raised when maximum error iterations are reached.

    This exception is raised when the maximum number of allowed errors has been
    reached. This includes both errors raised from the tool itself and errors
    when Yacana fails to call the tool correctly.

    Parameters
    ----------
    message : str
        Information about which specific iteration counter reached its maximum.

    Notes
    -----
    The error counters for ToolError and Yacana tool call errors are tracked
    separately. For example, if the maximum is set to 5, you could have 4 ToolError
    and 4 Yacana tool call errors (4 + 4 > 5) without raising this exception.
    The exception is only raised when one more error of either type occurs.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class ReachedTaskCompletion(Exception):
    """
    Exception raised when a task has been successfully completed.

    This exception is used to signal that the current task has reached its
    completion state. It is typically used to break out of processing loops
    or to indicate successful task termination.
    """

    def __init__(self):
        pass


class IllogicalConfiguration(Exception):
    """
    Exception raised when the framework is used in an incoherent way.

    This exception indicates that the framework has been configured or used
    in a way that doesn't make logical sense. The stacktrace message should
    provide details about the specific configuration issue.

    Parameters
    ----------
    message : str
        A description of the illogical configuration or usage.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class TaskCompletionRefusal(Exception):
    """
    Exception raised when the model refuses to complete a task.

    This exception is raised when the model explicitly refuses to complete
    the requested task, typically due to ethical, safety, or capability reasons.

    Parameters
    ----------
    message : str
        The reason for the task completion refusal.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class UnknownResponseFromLLM(Exception):
    """
    Exception raised when the model returns an unknown response.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class McpBadToolConfig(Exception):
    """
    Exception raised when the input shema from the MCP server is not valid.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class McpBadTransport(Exception):
    """
    Exception raised when the protocol used by the MCP server response was neither application/JSON or
    SSE and parsing as JSON failed.

    Parameters
    ----------
    message : str
        A descriptive message about the transport issue.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class McpResponseError(Exception):
    """
    Exception raised when the MCP server sent an error as part of the response.

    Parameters
    ----------
    message : str
        A descriptive message about the transport issue.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class McpServerNotYetInitialized(Exception):
    """
    Exception raised when the MCP server is not yet initialized.

    This exception is used to indicate that the MCP server has not been set up
    or is not ready to handle requests.

    Parameters
    ----------
    message : str
        A descriptive message indicating that the MCP server is not initialized.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class SpecializedTokenCountingError(Exception):
    """
    Exception raised when counting tokens using specialized methods fails.

    If this is raised then we fall back to counting tokens using regex approximations.

    Parameters
    ----------
    message : str
        A descriptive message indicating why token counting didn't work.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



from .task import Task
from .generic_agent import GenericAgent
from .open_ai_agent import OpenAiAgent
from .ollama_agent import OllamaAgent
from .agent import Agent
from .exceptions import MaxToolErrorIter, ToolError, IllogicalConfiguration, ReachedTaskCompletion
from .group_solve import EndChatMode, EndChat, GroupSolve
from .history import History, HistorySlot, SlotPosition
from .messages import (MessageRole, GenericMessage, Message, OpenAIUserMessage, OpenAITextMessage, OpenAIFunctionCallingMessage, OpenAiToolCallingMessage, OllamaToolCallingMessage, OpenAIStructuredOutputMessage, OllamaUserMessage,
                       OllamaTextMessage, OllamaStructuredOutputMessage)
from .logging_config import LoggerManager
from .model_settings import OllamaModelSettings
from .model_settings import OpenAiModelSettings
from .tool import Tool
from .tool import ToolType
from .mcp import Mcp



import unittest
import os
from tests.test_base import BaseAgentTest
from yacana import Task, Tool, ToolError, GenericMessage, GenericAgent, ToolType


def add(number_one: int, number_two: int) -> int:
    """Add two numbers together."""
    if type(number_one) is not int or type(number_two) is not int:
        raise ToolError("Both arguments must be integers.")
    return number_one + number_two


def subtract(number_one: int, number_two: int) -> int:
    """Subtract the second number from the first."""
    if type(number_one) is not int or type(number_two) is not int:
        raise ToolError("Both arguments must be integers.")
    return number_one - number_two


def multiply(number_one: int, number_two: int) -> int:
    """Multiply two numbers together."""
    if type(number_one) is not int or type(number_two) is not int:
        raise ToolError("Both arguments must be integers.")
    return number_one * number_two


def divide(number_one: int, number_two: int) -> int:
    """Divide the first number by the second."""
    if type(number_one) is not int or type(number_two) is not int:
        raise ToolError("Both arguments must be integers.")
    if number_two == 0:
        raise ToolError("Division by zero is not allowed.")
    return number_one // number_two


def get_weather(city: str) -> str:
    """Get the current weather for a city."""
    return f"The current real time weather from Weather.com in {city} is sunny with a high of 25°C."


def update_server() -> None:
    """Updates the server."""
    return None


def rollback_server() -> None:
    """Rollbacks the server."""
    return None


class TestToolCalling(BaseAgentTest):
    """Test tool calling capabilities of all agent types."""

    def setUp(self):
        """Set up test fixtures before each test."""
        super().setUp()
        # Create tools
        self.addition = Tool("Addition", "Add two integer numbers and returns the result.", add)
        self.subtraction = Tool("Subtraction", "Subtracts two integer numbers and returns the result.", subtract)
        self.multiplication = Tool("Multiplication", "Multiplies two integer numbers and returns the result.", multiply)
        self.division = Tool("Division", "Divides two integer numbers and returns the result.", divide)
        self.get_weather = Tool("Get_weather", "Calls a weather API and returns the current weather in the given city.", get_weather, shush=True)
        self.update_server = Tool("update_server", "Triggers a server update.", update_server)
        self.rollback_server = Tool("rollback_server", "Triggers a server rollback.", rollback_server)

    def test_basic_arithmetic(self):
        """Test basic arithmetic operations using tools."""

        def test_yacana_tool_calling(agent: GenericAgent):
            agent.history.clean()
            Tool.bulk_tool_type_update([self.addition, self.subtraction, self.multiplication, self.division], ToolType.YACANA)
            prompt = "Calculate 2+4-6*7 by decomposing the operations step by step and according to order of operations (PEMDAS/BODMAS). Use the provided tools. Do not make the math yourself. Only use the tools."
            message = Task(prompt, agent, tools=[self.addition, self.subtraction, self.multiplication, self.division]).solve()
            self.assertTrue(
                any(result in message.content for result in ["-36", "40"]),
                f"Expected either -36 or 40 in the result, got: {message.content}"
            )

        def test_openai_tool_calling(agent: GenericAgent):
            agent.history.clean()
            Tool.bulk_tool_type_update([self.addition, self.subtraction, self.multiplication, self.division], ToolType.OPENAI)
            Task(f"What's `2+4-6*7` ? You can only do one operation at a time. Don't worry you will be ask to continue with the operations later. Follow PEMDAS to solve correctly the equation. You are not allowed to guess the result of any operation. Wait for each tool result before continuing. Only call one tool at a time. When you believe you're finished output 'FINISH'.", agent, tools=[self.addition, self.subtraction, self.multiplication, self.division]).solve()
            while True:
                msg = Task(f"Continue solving the equation! You can only do one operation at a time. Don't worry you will be ask to continue with the operations later. Follow PEMDAS to solve correctly the equation. Only call one tool at a time. When you believe you're finished output 'FINISH'", agent, tools=[self.addition, self.subtraction, self.multiplication, self.division]).solve()
                if "finish" in msg.content.lower():
                    break

            one_before_last_message: GenericMessage = agent.history.get_message(len(agent.history.get_all_messages()) - 2)
            last_message: GenericMessage = agent.history.get_last_message()
            self.assertTrue(
                any(result in one_before_last_message.content for result in ["-36", "40"]) or
                any(result in last_message.content for result in ["-36", "40"]),
                f"Expected either -36 or 40 in the result, got: {one_before_last_message.content} or {last_message.content}"
            )

        # Test Ollama agent
        if self.run_ollama:
            test_yacana_tool_calling(self.ollama_agent)
            test_openai_tool_calling(self.ollama_agent)
        
        # Test OpenAI agent
        if self.run_openai:
            test_yacana_tool_calling(self.openai_agent)
            test_openai_tool_calling(self.openai_agent)
        
        # Test VLLM agent
        if self.run_vllm:
            test_yacana_tool_calling(self.vllm_agent)
            test_openai_tool_calling(self.vllm_agent)

        # Test LMStudio agent
        if self.run_lmstudio:
            test_yacana_tool_calling(self.lmstudio_agent)
            test_openai_tool_calling(self.lmstudio_agent)

    def test_weather_tool(self):
        """Test weather information tool."""
        prompt = "What's the weather in Paris?"
        
        # Test Ollama agent
        if self.run_ollama:
            self.get_weather.tool_type = ToolType.YACANA
            message = Task(prompt, self.ollama_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))

            self.ollama_agent.history.clean()

            self.get_weather.tool_type = ToolType.OPENAI
            message = Task(prompt, self.ollama_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))
        
        # Test OpenAI agent
        if self.run_openai:
            message = Task(prompt, self.openai_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))

            self.openai_agent.history.clean()

            self.get_weather.tool_type = ToolType.OPENAI
            message = Task(prompt, self.openai_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))
        
        # Test VLLM agent
        if self.run_vllm:
            message = Task(prompt, self.vllm_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))

            self.vllm_agent.history.clean()

            self.get_weather.tool_type = ToolType.OPENAI
            message = Task(prompt, self.vllm_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))

        # Test LMSTUDIO agent
        if self.run_lmstudio:
            message = Task(prompt, self.lmstudio_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))

            self.lmstudio_agent.history.clean()

            self.get_weather.tool_type = ToolType.OPENAI
            message = Task(prompt, self.lmstudio_agent, tools=[self.get_weather]).solve()
            self.assertTrue(all([
                "Paris" in message.content,
                "sunny" in message.content.lower(),
                "25" in message.content
            ]))

    def test_solo_empty_tool(self):
        """Test empty tool to check that no parameter on the called function doesn't break anything. Providing ONE tool only."""
        prompt = "Update the server."

        # Test Ollama agent
        if self.run_ollama:
            self.update_server.tool_type = ToolType.YACANA
            Task(prompt, self.ollama_agent, tools=[self.update_server]).solve()

            self.ollama_agent.history.clean()

            self.update_server.tool_type = ToolType.OPENAI
            Task(prompt, self.ollama_agent, tools=[self.update_server]).solve()

        # Test LMStudio agent
        if self.run_lmstudio:
            self.update_server.tool_type = ToolType.YACANA
            Task(prompt, self.lmstudio_agent, tools=[self.update_server]).solve()

            self.lmstudio_agent.history.clean()

            self.update_server.tool_type = ToolType.OPENAI
            Task(prompt, self.lmstudio_agent, tools=[self.update_server]).solve()


    def test_multi_empty_tools(self):
        """Test empty tool to check that no parameter on the called function doesn't break anything. Providing multiple tools."""
        prompt = "Update the server."

        # Test Ollama agent
        if self.run_ollama:
            self.update_server.tool_type = ToolType.YACANA
            self.rollback_server.tool_type = ToolType.YACANA
            Task(prompt, self.ollama_agent, tools=[self.update_server, self.rollback_server]).solve()

            self.ollama_agent.history.clean()

            self.update_server.tool_type = ToolType.OPENAI
            self.rollback_server.tool_type = ToolType.OPENAI
            Task(prompt, self.ollama_agent, tools=[self.update_server, self.rollback_server]).solve()

        # Test LMStudio agent
        if self.run_lmstudio:
            self.update_server.tool_type = ToolType.YACANA
            self.rollback_server.tool_type = ToolType.YACANA
            Task(prompt, self.lmstudio_agent, tools=[self.update_server, self.rollback_server]).solve()

            self.lmstudio_agent.history.clean()

            self.update_server.tool_type = ToolType.OPENAI
            self.rollback_server.tool_type = ToolType.OPENAI
            Task(prompt, self.lmstudio_agent, tools=[self.update_server, self.rollback_server]).solve()


    @classmethod
    def setUpClass(cls):
        """Set up test fixtures before running tests."""
        super().setUpClass()
        
        # Get which agents to run from environment variables
        cls.run_ollama = os.getenv('TEST_OLLAMA', 'true').lower() == 'true'
        cls.run_openai = os.getenv('TEST_OPENAI', 'true').lower() == 'true'
        cls.run_vllm = os.getenv('TEST_VLLM', 'true').lower() == 'true'
        cls.run_vllm = os.getenv('TEST_LMSTUDIO', 'true').lower() == 'true'

if __name__ == '__main__':
    unittest.main()



import copy
import logging
import uuid
from typing import List, Type, Callable, Dict
from pydantic import BaseModel

from .generic_agent import GenericAgent, GenericMessage
from .exceptions import IllogicalConfiguration
from .history import History
from .logging_config import LoggerManager
from .tool import Tool
from .history import Message

LoggerManager.set_library_log_level("httpx", "WARNING")


class Task:
    """
    A class representing a task to be solved by an LLM agent.

    The best approach to use an LLM is to define a task with a clear objective.
    Then assign an LLM to this task so that it can try to solve it. You can also add Tools
    so that when the LLM starts solving it gets some tools relevant to the current task.
    This means that tools are not specific to an agent but to the task at hand. This allows
    more flexibility by producing less confusion to the LLM as it gets access to the tools
    it needs only when faced with a task that is related.

    Parameters
    ----------
    prompt : str
        The task to solve. It is the prompt given to the assigned LLM.
    agent : GenericAgent
        The agent assigned to this task.
    json_output : bool, optional
        If True, will force the LLM to answer as JSON. Defaults to False.
    structured_output : Type[BaseModel] | None, optional
        The expected structured output type for the task. If provided, the LLM's response
        will be validated against this type. Defaults to None.
    tools : List[Tool], optional
        A list of tools that the LLM will get access to when trying to solve this task.
        Defaults to an empty list.
    medias : List[str] | None, optional
        An optional list of paths pointing to images on the filesystem. Defaults to None.
    llm_stops_by_itself : bool, optional
        Only useful when the task is part of a GroupSolve(). Signals the assigned LLM
        that it will have to stop talking by its own means. Defaults to False.
    use_self_reflection : bool, optional
        Only useful when the task is part of a GroupSolve(). Allows keeping the self
        reflection process done by the LLM in the next GS iteration. Defaults to False.
    forget : bool, optional
        When True, the Agent won't remember this task after completion. Useful for
        routing purposes. Defaults to False.
    streaming_callback : Callable | None, optional
        Optional callback for streaming responses. Defaults to None.
    runtime_config : Dict | None, optional
        Optional runtime configuration for the task. Defaults to None.
    tags : List[str] | None, optional
        Optional list of tags that will be added to all message(s) corresponding to this task. Defaults to None.

    Attributes
    ----------
    prompt : str
        The task to solve. It is the prompt given to the assigned LLM.
    agent : GenericAgent
        The agent assigned to this task.
    json_output : bool
        If True, will force the LLM to answer as JSON.
    structured_output : Type[BaseModel] | None
        The expected structured output type for the task.
    tools : List[Tool]
        A list of tools that the LLM will get access to when trying to solve this task.
    medias : List[str] | None
        An optional list of paths pointing to images on the filesystem.
    llm_stops_by_itself : bool
        Only useful when the task is part of a GroupSolve(). Signals the assigned LLM
        that it will have to stop talking by its own means.
    use_self_reflection : bool
        Only useful when the task is part of a GroupSolve(). Allows keeping the self
        reflection process done by the LLM in the next GS iteration.
    forget : bool
        When True, the Agent won't remember this task after completion. Useful for
        routing purposes.
    streaming_callback : Callable | None
        Optional callback for streaming responses.
    runtime_config : Dict | None
        Optional runtime configuration for the task.
    tags : List[str] | None
        Optional list of tags that will be added to all message(s) corresponding to this task.

    Raises
    ------
    IllogicalConfiguration
        If both tools and structured_output are provided, or if both streaming_callback
        and structured_output are provided.
    """

    def __init__(self, prompt: str, agent: GenericAgent, json_output=False, structured_output: Type[BaseModel] | None = None, tools: List[Tool] = None,
                 medias: List[str] | None = None, llm_stops_by_itself: bool = False, use_self_reflection=False, forget=False, streaming_callback: Callable | None = None, runtime_config: Dict | None = None, tags: List[str] = None) -> None:
        self.prompt: str = prompt
        self.agent: GenericAgent = agent
        self.json_output: bool = json_output
        self.structured_output: Type[BaseModel] | None = structured_output
        self.tools: List[Tool] = tools if tools is not None else []
        self.llm_stops_by_itself: bool = llm_stops_by_itself
        self.use_self_reflection: bool = use_self_reflection
        self.forget: bool = forget
        self.medias: List[str] | None = medias
        self._uuid: str = str(uuid.uuid4())
        self.streaming_callback: Callable | None = streaming_callback
        self.runtime_config = runtime_config if runtime_config is not None else {}
        self.tags: List[str] = tags if tags is not None else []

        if len(self.tools) > 0 and self.structured_output is not None:
            raise IllogicalConfiguration("You can't have tools and structured_output at the same time. The tool output will be considered the LLM output hence not using the structured output.")

        if self.streaming_callback is not None and self.structured_output is not None:
            raise IllogicalConfiguration("You can't have streaming_callback and structured_output at the same time. Having incomplete JSON is useless.")

        self._check_tools_are_same_type()
        self._check_tool_names_are_unique()

        # Only used when @forget is True
        self._initial_history: History | None = None

    @property
    def uuid(self) -> str:
        """
        Get the unique identifier for this task.

        Returns
        -------
        str
            A unique task identifier.
        """
        return self._uuid

    def _check_tools_are_same_type(self):
        """
        All tools must be of the same type. We can't mix tools because they are not proposed in the same way to the LLM.
        """
        if len(self.tools) > 0:
            first_tool_type = self.tools[0].tool_type
            for tool in self.tools:
                if tool.tool_type != first_tool_type:
                    raise IllogicalConfiguration("All tools must be of the same type. Mixing tool types is not allowed. Use ToolType.YACANA or ToolType.OPENAI to specify the type of tool execution you want to use.")

    def _check_tool_names_are_unique(self):
        """
        Check that all local tools have unique names and that MCP tools do not conflict with local tool names.
        """
        local_tool_names: List[str] = [tool.tool_name for tool in self.tools if not tool.is_mcp]

        if len(set(local_tool_names)) != len(local_tool_names):
            raise IllogicalConfiguration("All local tools must have unique names. Found duplicates in the task's tool list.")

        unique_mcp_tools: List[Tool] = []
        mcp_tools: List[Tool] = [tool for tool in self.tools if tool.is_mcp]

        for mcp_tool in mcp_tools:
            if mcp_tool.tool_name not in local_tool_names:
                unique_mcp_tools.append(mcp_tool)
            else:
                logging.warning(
                    "Tool '%s' is a MCP tool but its name '%s' is already used by a tool defined at Task "
                    "level. Local task tools take precedence upon remote MCP tools for security reasons. You should either change your local tool name or use the .forget_tool() method from the Mcp class. For now the remote tool will no be available.",
                    mcp_tool.tool_name,
                    mcp_tool.tool_name
                )

        self.tools = [tool for tool in self.tools if not tool.is_mcp] + unique_mcp_tools

    def add_tool(self, tool: Tool) -> None:
        """
        Add a tool to the list of tools available for this task.

        Parameters
        ----------
        tool : Tool
            The tool to add to the task's tool list.
        """
        self.tools.append(tool)

    def _save_history_state(self) -> None:
        """Save the current state of the agent's history."""
        if self.forget:
            self._initial_history = copy.deepcopy(self.agent.history)

    def _restore_history_state(self) -> None:
        """Restore the agent's history to its initial state."""
        if self.forget and self._initial_history is not None:
            self.agent.history = self._initial_history

    def solve(self) -> Message:
        """
        Execute the task using the assigned LLM agent.

        This method will call the assigned LLM to perform inference on the task's prompt.
        If tools are available, the LLM may use them, potentially making multiple calls
        to the inference server.

        Returns
        -------
        GenericMessage | None
            The last message from the LLM.

        Raises
        ------
        MaxToolErrorIter
            If tool errors exceed the limit.
        """
        self._save_history_state()
        answer: GenericMessage = self.agent._interact(self.prompt, self.tools, self.json_output, self.structured_output, self.medias, self.streaming_callback, self.runtime_config, self.tags)
        self._restore_history_state()
        return answer




import unittest
import os
from tests.test_base import BaseAgentTest
from yacana import Task
from yacana.exceptions import IllogicalConfiguration

class TestStreaming(BaseAgentTest):
    """Test streaming capabilities of all agent types."""

    def test_basic_streaming(self):
        """Test basic streaming functionality."""
        prompt = "Count from 1 to 5 (no additionnal text, numbers only):"
        expected_numbers = {"1", "2", "3", "4", "5"}
        received_numbers = set()
        
        def streaming_callback(chunk: str):
            self.assertIsInstance(chunk, str)
            for num in expected_numbers:
                if num in chunk:
                    received_numbers.add(num)
        
        # Test Ollama agent
        if self.run_ollama:
            received_numbers.clear()
            message = Task(prompt, self.ollama_agent, streaming_callback=streaming_callback).solve()
            self.assertEqual(received_numbers, expected_numbers)
        
        # Test OpenAI agent
        if self.run_openai:
            received_numbers.clear()
            message = Task(prompt, self.openai_agent, streaming_callback=streaming_callback).solve()
            self.assertEqual(received_numbers, expected_numbers)
        
        # Test VLLM agent
        if self.run_vllm:
            received_numbers.clear()
            message = Task(prompt, self.vllm_agent, streaming_callback=streaming_callback).solve()
            self.assertEqual(received_numbers, expected_numbers)

    def test_streaming_with_media(self):
        """Test streaming with image input."""
        image_path = self.get_test_image_path("burger.jpg")
        prompt = "Describe this image in one sentence:"
        received_chunks = []
        
        def streaming_callback(chunk: str):
            self.assertIsInstance(chunk, str)
            received_chunks.append(chunk)
        
        # Test Ollama agent
        if self.run_ollama:
            received_chunks.clear()
            message = Task(prompt, self.ollama_vision_agent, medias=[image_path], streaming_callback=streaming_callback).solve()
            full_response = "".join(received_chunks)
            self.assertGreater(len(full_response), 0)
            self.assertTrue(
                any(keyword in full_response.lower() for keyword in ['food', 'burger', 'meal', 'dish', 'eating']),
                "Response should mention food-related content"
            )
        
        # Test OpenAI agent
        if self.run_openai:
            received_chunks.clear()
            message = Task(prompt, self.openai_agent, medias=[image_path], streaming_callback=streaming_callback).solve()
            full_response = "".join(received_chunks)
            self.assertGreater(len(full_response), 0)
            self.assertTrue(
                any(keyword in full_response.lower() for keyword in ['food', 'burger', 'meal', 'dish', 'eating']),
                "Response should mention food-related content"
            )
        
        # Test VLLM agent
        #if self.run_vllm:
        #    received_chunks.clear()
        #    message = Task(prompt, self.vllm_agent, medias=[image_path], streaming_callback=streaming_callback).solve()
        #    full_response = "".join(received_chunks)
        #    self.assertGreater(len(full_response), 0)
        #    self.assertTrue(
        #        any(keyword in full_response.lower() for keyword in ['food', 'burger', 'meal', 'dish', 'eating']),
        #        "Response should mention food-related content"
        #    )

    def test_streaming_with_structured_output(self):
        """Test that using both streaming and structured output raises IllogicalConfiguration."""
        from pydantic import BaseModel

        class Person(BaseModel):
            name: str
            age: int

        prompt = "Generate a person's information with name and age:"
        
        def streaming_callback(chunk: str):
            pass  # This should never be called
        
        # Test Ollama agent
        if self.run_ollama:
            with self.assertRaises(IllogicalConfiguration):
                Task(prompt, self.ollama_agent, structured_output=Person, streaming_callback=streaming_callback).solve()
        
        # Test OpenAI agent
        if self.run_openai:
            with self.assertRaises(IllogicalConfiguration):
                Task(prompt, self.openai_agent, structured_output=Person, streaming_callback=streaming_callback).solve()
        
        # Test VLLM agent
        if self.run_vllm:
            with self.assertRaises(IllogicalConfiguration):
                Task(prompt, self.vllm_agent, structured_output=Person, streaming_callback=streaming_callback).solve()

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures before running tests."""
        super().setUpClass()
        
        # Get which agents to run from environment variables
        cls.run_ollama = os.getenv('TEST_OLLAMA', 'true').lower() == 'true'
        cls.run_openai = os.getenv('TEST_OPENAI', 'true').lower() == 'true'
        cls.run_vllm = os.getenv('TEST_VLLM', 'true').lower() == 'true'

if __name__ == '__main__':
    unittest.main() 


from typing import Callable
from pydantic import create_model
import inspect

from .exceptions import IllogicalConfiguration


def function_to_json_with_pydantic(tool_name: str, description: str, func: Callable) -> dict:
    """Convert a Python function to a JSON description with an exact match format."""

    # We keep this for later use, but we don't need it for now as it's not the function name that is used but the tool name given in the Tool constructor.
    func_name = func.__name__

    # Extract parameters and annotations
    signature = inspect.signature(func)
    annotations = {name: param.annotation for name, param in signature.parameters.items()}

    # Checking annotations existence
    for name, annotation in annotations.items():
        if annotation == inspect._empty:
            raise IllogicalConfiguration(
                "A function used as a Tool must have all its parameters being duck typed in its prototype."
            )

    required_params = [
        name
        for name, param in signature.parameters.items()
        if param.default == inspect.Parameter.empty
    ]

    # Create a Pydantic model dynamically for the function parameters
    PydanticModel = create_model(
        func_name.capitalize() + "Params",
        **{
            name: (annotation, ...)
            if name in required_params
            else (annotation, None)
            for name, annotation in annotations.items()
        }
    )

    # Extract properties and required fields from Pydantic model schema
    schema = PydanticModel.schema()
    properties = schema.get("properties", {})
    required = schema.get("required", [])

    # Map types to match desired format
    def map_pydantic_to_json_schema(prop):
        json_type_mapping = {
            "integer": "number",
            "string": "string",
            "array": "array",
            "object": "object",
            "boolean": "boolean",
            "number": "number",
        }
        prop_type = prop.get("type", "string")
        mapped_type = json_type_mapping.get(prop_type, "string")
        result = {"type": mapped_type}
        if mapped_type == "array" and "items" in prop:
            result["items"] = map_pydantic_to_json_schema(prop["items"])
        return result

    # Transform properties to match your JSON structure
    transformed_properties = {
        name: map_pydantic_to_json_schema(details)
        for name, details in properties.items()
    }

    # Construct JSON structure
    func_json = {
        "type": "function",
        "function": {
            "name": tool_name,
            "description": description,
            "parameters": {
                "type": "object",
                "properties": transformed_properties,
                "required": required,
                "additionalProperties": False
            },
            "strict": True
        }
    }

    return func_json


def input_schema_to_function_json(tool_name: str, description: str, input_schema: dict) -> dict:
    """
    Convert a JSON input schema to the structured function calling format.

    Args:
        tool_name (str): Name of the tool/function.
        description (str): Description of the tool/function.
        input_schema (dict): The schema describing parameters, e.g. {"type": "object", "properties": ..., "required": ...}.

    Returns:
        dict: JSON structure in the format expected for LLM function calls.
    """

    def map_json_schema_type(prop: dict) -> dict:
        """Map JSON Schema types to expected format."""
        json_type_mapping = {
            "integer": "number",
            "string": "string",
            "array": "array",
            "object": "object",
            "boolean": "boolean",
            "number": "number",
        }
        prop_type = prop.get("type", "string")
        mapped_type = json_type_mapping.get(prop_type, "string")

        result = {"type": mapped_type}

        if mapped_type == "array" and "items" in prop:
            result["items"] = map_json_schema_type(prop["items"])

        return result

    properties = input_schema.get("properties", {})
    required = input_schema.get("required", [])

    # Map and transform properties
    transformed_properties = {
        name: map_json_schema_type(prop)
        for name, prop in properties.items()
    }

    return {
        "type": "function",
        "function": {
            "name": tool_name,
            "description": description,
            "parameters": {
                "type": "object",
                "properties": transformed_properties,
                "required": required,
                "additionalProperties": False
            },
            "strict": True
        }
    }



import logging


class ColoredFormatter(logging.Formatter):
    """
    A custom formatter that adds color to log messages based on their level and content.

    This formatter uses ANSI escape codes to colorize log messages. Special formatting is applied
    for different types of messages like prompts, AI responses, and tool responses.

    Attributes
    ----------
    COLORS : dict
        Dictionary mapping log levels to their corresponding ANSI color codes.
    RESET : str
        ANSI code to reset text color.
    """
    # ANSI escape codes for colors
    COLORS = {
        'DEBUG': '\033[94m',  # Blue
        'INFO': '\033[92m',  # Green
        'WARNING': '\033[93m',  # Yellow
        'ERROR': '\033[91m',  # Red
        'CRITICAL': '\033[96m'  # CYAN
    }
    RESET = '\033[0m'

    # Magenta '\033[95m'

    def format(self, record: logging.LogRecord) -> str:
        """
        Format the log record with appropriate colors.

        Parameters
        ----------
        record : logging.LogRecord
            The log record to format.

        Returns
        -------
        str
            The formatted log message with appropriate color codes.
        """
        if record.levelname == "INFO" and record.msg.startswith("[PROMPT]"):
            log_color: str = '\033[92m'  # Green
        elif record.levelname == "INFO" and record.msg.startswith("[AI_RESPONSE]"):
            log_color: str = '\033[95m'  # Magenta
        elif record.levelname == "INFO" and record.msg.startswith("[TOOL_RESPONSE]"):
            log_color: str = '\033[96m'  # Magenta
        else:
            log_color: str = self.COLORS.get(record.levelname, self.RESET)
        log_message: str = super().format(record)
        return f"\n{log_color}{log_message}{self.RESET}"


class LoggerManager:
    """
    Manages logging configuration for the application.

    This class provides static methods to configure logging levels and handlers.
    It allows setting different log levels for the main application and specific libraries.
    Defaults to INFO level and includes colored output for different types of messages.

    Attributes
    ----------
    LOG_LEVEL : str
        Default log level for the application.
    AVAILABLE_LOG_LEVELS : list[str]
        List of valid log levels that can be set.
    """
    LOG_LEVEL = "INFO"
    AVAILABLE_LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", None]

    @staticmethod
    def setup_logging(log_level: str | None = None) -> None:
        """
        Initialize logging configuration for the application.

        This method sets up the root logger with a colored formatter and configures
        the logging level. It should not be called directly unless you know what you
        are doing. Use set_log_level() instead for normal usage.

        Parameters
        ----------
        log_level : str | None, optional
            The desired log level. Must be one of: "DEBUG", "INFO", "WARNING",
            "ERROR", "CRITICAL", or None. If None, uses the default LOG_LEVEL.
            Defaults to None.

        Notes
        -----
        This method is automatically called when the module is imported.
        """
        if log_level is None:
            log_level = LoggerManager.LOG_LEVEL
        else:
            log_level = log_level.upper()

        if log_level not in LoggerManager.AVAILABLE_LOG_LEVELS:
            log_level = "INFO"
            print("WARNING:root:Invalid value given to LOG_LEVEL. Defaulting to level: INFO")

        # Create and configure the handler with the ColoredFormatter
        handler = logging.StreamHandler()
        formatter = ColoredFormatter('%(levelname)s: %(message)s')
        handler.setFormatter(formatter)

        # Configure the root logger
        root_logger = logging.getLogger()
        root_logger.setLevel(getattr(logging, log_level))
        root_logger.handlers = [handler]  # Replace default handlers with our colored handler

    @staticmethod
    def set_log_level(log_level: str | None) -> None:
        """
        Set the logging level for the application.

        This method allows changing the logging level at runtime. All log levels have
        specific colors, and for INFO level, the color depends on whether it's a user
        prompt or an AI answer. Setting log_level to None will disable logging.

        Parameters
        ----------
        log_level : str | None
            The desired log level. Must be one of: "DEBUG", "INFO", "WARNING",
            "ERROR", "CRITICAL", or None. If None, logging will be disabled.

        Notes
        -----
        If an invalid log level is provided, the current level will be maintained.
        """
        if log_level is None:
            # Disable logging if the log level is None
            logging.disable(logging.CRITICAL)
        elif log_level in LoggerManager.AVAILABLE_LOG_LEVELS:
            logging.getLogger().setLevel(getattr(logging, log_level))
            logging.disable(logging.NOTSET)  # Re-enable logging
        else:
            print(
                f"WARNING:root:Invalid value given to LOG_LEVEL. Keeping current level: {logging.getLevelName(logging.getLogger().level)}")

    @staticmethod
    def set_library_log_level(library_name: str, log_level: str) -> None:
        """
        Set the logging level for a specific Python library.

        This is useful to control logging from external libraries independently
        from the main application logging level.

        Parameters
        ----------
        library_name : str
            The name of the target library.
        log_level : str
            The desired log level. Must be one of: "DEBUG", "INFO", "WARNING",
            "ERROR", "CRITICAL", or None.

        Notes
        -----
        If an invalid log level is provided, a warning will be printed but no changes
        will be made to the library's logging configuration.
        """
        if log_level in LoggerManager.AVAILABLE_LOG_LEVELS:
            library_logger = logging.getLogger(library_name)
            library_logger.setLevel(getattr(logging, log_level))
        else:
            print(f"WARNING:root:Invalid log level {log_level} for library {library_name}.")


# Automatically set up logging when the class is first imported
LoggerManager.setup_logging()



import json
import logging
from json import JSONDecodeError
from typing import List, Type, T, Dict, Callable
from pydantic import BaseModel

from .exceptions import MaxToolErrorIter, ToolError, UnknownResponseFromLLM
from .messages import OpenAiToolCallingMessage, OpenAIFunctionCallingMessage, OllamaToolCallingMessage
from .tool import Tool
from .base_tool_caller import BaseToolCaller
from .utils import AgentType


class OpenAiToolCaller(BaseToolCaller):

    def propose_tool(self, task: str, tools: List[Tool], json_output: bool, structured_output: Type[BaseModel] | None, medias: List[str] | None, streaming_callback: Callable | None = None, task_runtime_config: Dict | None = None, tags: List[str] | None = None):
        self.propose_tools(task, tools, json_output, structured_output, medias, streaming_callback, task_runtime_config, tags)

    def propose_tools(self, task: str, tools: List[Tool], json_output: bool, structured_output: Type[BaseModel] | None, medias: List[str] | None, streaming_callback: Callable | None = None, task_runtime_config: Dict | None = None, tags: List[str] | None = None):
        self.agent._chat(self.agent.history, task, medias=medias, json_output=json_output, structured_output=structured_output, tools=tools)
        if isinstance(self.agent.history.get_last_message(), OpenAIFunctionCallingMessage):
            for tool_call in self.agent.history.get_last_message().tool_calls:
                tool = next((tool for tool in tools if tool.tool_name == tool_call.name), None)
                if tool is None:
                    raise ValueError(f"Tool {tool_call.name} not found in tools list")
                logging.debug("Found tool: %s", tool.tool_name)
                func_arguments: dict = {}
                try:
                    if isinstance(tool_call.arguments, dict):
                        func_arguments = tool_call.arguments
                    else:
                        func_arguments: dict = json.loads(tool_call.arguments)
                        if not isinstance(func_arguments, dict): # In case of multi JSON encoding layers, arguments mays still not be a dict
                            raise ValueError(f"Still not a valid dict after doing json.loads on arguments.")
                except Exception as e:
                    raise UnknownResponseFromLLM(f"Tool {tool_call.name} arguments are not a valid dict object: `{tool_call.arguments}`. Type: {type(tool_call.arguments)}. Original error: {e}.")
                tool_output: str = self._call_openai_tool(tool, func_arguments)
                if self.agent._agent_type == AgentType.OPENAI:
                    self.agent.history.add_message(OpenAiToolCallingMessage(tool_output, tool_call.call_id, tags=self.agent._tags))
                elif self.agent._agent_type == AgentType.OLLAMA:
                    self.agent.history.add_message(OllamaToolCallingMessage(tool_output, tool_call.name, tags=self.agent._tags))

            logging.info(f"[PROMPT][To: {self.agent.name}]: Retrying with original task and tools answer: '{task}'")
            self.agent._chat(self.agent.history, None, medias=medias, json_output=json_output, structured_output=structured_output, streaming_callback=streaming_callback)

    def _call_openai_tool(self, tool: Tool, function_args: Dict) -> str:
        """
        Executes a tool call and handles any errors that occur.

        Parameters
        ----------
        tool : Tool
            The tool to execute.
        function_args : Dict
            The arguments to pass to the tool function.

        Returns
        -------
        str
            The output from the tool execution.

        Raises
        ------
        MaxToolErrorIter
            If too many errors occur during tool execution.
        """
        max_call_error: int = tool.max_call_error
        max_custom_error: int = tool.max_custom_error
        tool_output: str = ""

        while True:
            try:
                if tool.is_mcp:
                    tool_output: str = tool.function_ref(tool_name=tool.tool_name, arguments=function_args)
                else:
                    tool_output: str = tool.function_ref(**function_args)
                if tool_output is None:
                    tool_output = f"Tool returned 'None' so LLM won't be asked to reflect on the tool result."
                else:
                    tool_output = str(tool_output)
                logging.info(f"[TOOL_RESPONSE][{tool.tool_name}]: {tool_output}\n")
                break
            except (ToolError, TypeError, JSONDecodeError) as e:
                if type(e) is ToolError or type(e) is JSONDecodeError:
                    logging.warning(f"Tool '{tool.tool_name}' raised an error\n")
                    max_custom_error -= 1
                    tool_output = e.message
                elif type(e) is TypeError:
                    logging.warning(f"Yacana failed to call tool '{tool.tool_name}' correctly based on the LLM output\n")
                    tool_output = str(e)
                    max_call_error -= 1

                if max_custom_error < 0:
                    raise MaxToolErrorIter(
                        f"Too many errors were raise by the tool '{tool.tool_name}'. Stopping after {tool.max_custom_error} errors. You can change the maximum errors a tool can raise in the Tool constructor with @max_custom_error.")
                if max_call_error < 0:
                    raise MaxToolErrorIter(
                        f"Too many errors occurred while trying to call the python function by Yacana (tool name: {tool.tool_name}). Stopping after {tool.max_call_error} errors. You can change the maximum call error in the Tool constructor with @max_call_error.")
                self.agent._chat(self.agent.history, f"The tool returned an error: `{tool_output}`\nUsing this error message, fix the JSON you generated.")
        return tool_output



import copy
import json
import uuid
from datetime import datetime
from enum import Enum, unique
from typing import List, Dict, T, Sequence
from typing_extensions import Self
import logging

from .TokenCount import count_tokens_using_huggingface, HFMessage
from .exceptions import SpecializedTokenCountingError, IllogicalConfiguration
from .messages import GenericMessage, Message, MessageRole


@unique
class SlotPosition(Enum):
    """
    <b>ENUM:</b> The position of a slot in the history. This is only a syntactic sugar to make the code more readable.

    Attributes
    ----------
    BOTTOM : int
        The slot is at the bottom of the history.
    TOP : int
        The slot is at the top of the history.
    """
    BOTTOM = -1
    TOP = -2


class HistorySlot:
    """
    A slot is a container for messages. It can contain one or more messages.
    Most of the time it will only contain one message but when using `n=2` or`n=x` in the OpenAI API, it will contain multiple variations hence multiple messages.

    Parameters
    ----------
    messages : List[GenericMessage], optional
        A list of messages. Each message is a variation of the main message (defined by the
        @main_message_index parameter).
    raw_llm_json : str, optional
        The raw LLM JSON response for the slot. This is the raw JSON from the inference server.
        When using OpenAI this may contain more than one message hence the slot system acts as a
        container for the messages.
    **kwargs
        Additional keyword arguments including:
        id : str, optional
            The unique identifier for the slot.
        creation_time : int, optional
            The timestamp when the slot was created.

    Attributes
    ----------
    id : str
        The unique identifier for the slot.
    creation_time : int
        The timestamp when the slot was created.
    messages : List[GenericMessage]
        List of messages in the slot.
    raw_llm_json : str | None
        The raw LLM JSON response for the slot.
    main_message_index : int
        The index of the main message in the slot.
    """

    def __init__(self, messages: List[GenericMessage] = None, raw_llm_json: str = None, **kwargs):
        self.id = str(kwargs.get('id', uuid.uuid4()))
        self.creation_time: int = int(kwargs.get('creation_time', datetime.now().timestamp()))
        self.messages: List[GenericMessage] = [] if messages is None else messages
        self.raw_llm_json: str = raw_llm_json
        self.main_message_index: int = 0

    def set_main_message_index(self, message_index: int) -> None:
        """
        A slot can contain any number of concurrent message. But only one can be the main slot message and actually be part of the History.
        This method sets the index of the main message within the list of available messages in the slot.

        Parameters
        ----------
        message_index : int
            The index of the message to select as the main message.

        Raises
        ------
        IndexError
            If the message index is greater than the number of messages in the slot.
        """
        if message_index >= len(self.messages):
            raise IndexError("Index out of range: The message index is greater than the number of messages in the slot.")
        self.main_message_index = message_index

    def get_main_message_index(self) -> int:
        """
        Returns the index of the main message in the slot.

        Returns
        -------
        int
            The index of the currently selected message.
        """
        return self.main_message_index

    def add_message(self, message: GenericMessage):
        """
        Adds a new message to the slot.

        Parameters
        ----------
        message : GenericMessage
            The message to add to the slot.
        """
        self.messages.append(message)

    def get_message(self, message_index: int | None = None) -> GenericMessage:
        """
        Returns the main message of the slot or the one at the given index if index is provided.

        Parameters
        ----------
        message_index : int | None, optional
            The index of the message to return. If None, returns the currently selected message.

        Returns
        -------
        GenericMessage
            The requested message.

        Raises
        ------
        IndexError
            If the message index is greater than the number of messages in the slot.
        IllogicalConfiguration
            An HistorySlot should not be empty (no messages).
        """
        if message_index is None:
            return self.messages[self.main_message_index]
        if message_index >= len(self.messages):
            raise IndexError("Index out of range: The message index is greater than the number of messages in the slot.")
        else:
            return self.messages[message_index]

    def get_all_messages(self) -> List[GenericMessage]:
        """
        Returns all the messages in the slot.

        Returns
        -------
        List[GenericMessage]
            All messages in the slot.
        """
        return self.messages

    def set_raw_llm_json(self, raw_llm_json: str) -> None:
        """
        Sets the raw LLM JSON response for the slot.
        This is the raw JSON from the inference server. When using OpenAI this may contain more than one message hence the slot system acts as a container for the messages.

        Parameters
        ----------
        raw_llm_json : str
            The raw JSON response from the LLM.
        """
        self.raw_llm_json = raw_llm_json

    def _delete_message_by_index(self, message_index: int) -> None:
        """
        Deletes a message from the slot by index. An HistorySlot should NOT be empty (no messages). Use this method at your own risk.
        To delete a message from a slot, delete it using the `delete_message` method from the History class instead.
        This way, if the slot ends being empty it will be cleaned by the History class.

        Parameters
        ----------
        message_index : int
            The index of the message to delete.

        Raises
        ------
        IndexError
            If the message index is greater than the number of messages in the slot,
            or if trying to delete the last message in the slot.
        """
        if message_index >= len(self.messages):
            raise IndexError("Index out of range: The message index is greater than the number of messages in the slot.")
        
        if len(self.messages) <= 1:
            raise IndexError("Cannot delete the last message in a slot. Delete the slot from the history instead.")
        
        self.messages.pop(message_index)
        
        # Always reset to the first message for simplicity
        self.main_message_index = 0
        logging.debug("Main message index reset to 0 after message deletion")

    def _delete_message_by_id(self, message_id: str) -> None:
        """
        Deletes a message from the slot by id. An HistorySlot should NOT be empty (no messages). Use this method at your own risk.
        To delete a message from a slot, delete it using the `delete_message` method from the History class instead.
        This way, if the slot ends being empty it will be cleaned by the History class.

        Parameters
        ----------
        message_id : str
            The ID of the message to delete.

        Raises
        ------
        IndexError
            If trying to delete the last message in the slot.
        """
        for i, message in enumerate(self.messages):
            if message.id == message_id:
                self._delete_message_by_index(i)
                break

    def keep_only_selected_message(self):
        """
        Keeps only the currently selected message in the slot and deletes all the others.
        If there's only one message, this method does nothing.

        Raises
        ------
        IndexError
            If there are no messages in the slot.
        """
        if len(self.messages) == 0:
            raise IndexError("Cannot operate on an empty slot.")
            
        if len(self.messages) == 1:
            logging.debug("Keeping only selected message: Only one message in slot. Nothing to do.")
            return
            
        # Store the main message
        main_message = self.messages[self.main_message_index]
        
        # Clear all messages
        self.messages.clear()
        
        # Add back only the main message
        self.messages.append(main_message)
        self.main_message_index = 0
        logging.debug("Main message index reset to 0 after keeping only the selected message.")

    @staticmethod
    def create_instance(members: Dict):
        """
        Creates an instance of the HistorySlot class from a dictionary.
        Mainly used to import the object from a file.

        Parameters
        ----------
        members : Dict
            Dictionary containing the slot data.

        Returns
        -------
        HistorySlot
            A new instance of HistorySlot.
        """
        members["messages"] = [GenericMessage.create_instance(message) for message in members["messages"] if message is not None]
        return HistorySlot(**members)

    def _export(self) -> Dict:
        """
        Returns the slot as a dictionary.
        Mainly used to export the object to a file.

        Returns
        -------
        Dict
            A dictionary representation of the slot.
        """
        members = self.__dict__.copy()
        members["type"] = self.__class__.__name__
        members["messages"] = [message._export() for message in self.messages if message is not None]
        return members


class History:
    """
    Container for an alternation of Messages representing a conversation between the user and an LLM.
    To be precise, the history is a list of slots and not actual messages. Each slot contains at least one or more messages.
    This class does its best to hide the HistorySlot implementation. Meaning that many methods allows you to deal with the messages directly, but under the hood it always manages the slot wrapper.

    Parameters
    ----------
    llm_model_name : str | None, optional
        The name of the LLM model. Used to count tokens more accurately when using an OpenAi model. Will use Tiktoken under the hood.
    hugging_face_details: HuggingFaceDetails | None, optional
        Details for Hugging Face models, including repo name and access token. Used to count tokens more accurately when using an HuggingFace model. Will use the transformers library under the hood.
    **kwargs: Any
        Additional keyword arguments including:
        slots : List[HistorySlot], optional
            List of history slots.
        _checkpoints : Dict[str, list[HistorySlot]], optional
            Dictionary of checkpoints for the history.

    Attributes
    ----------
    slots : List[HistorySlot]
        List of history slots.
    _checkpoints : Dict[str, list[HistorySlot]]
        Dictionary of checkpoints for the history.
    llm_model_name : str | None
        The name of the LLM model used to count tokens when using an OpenAi model.
    """

    def __init__(self, llm_model_name: str | None = None, **kwargs) -> None:
        self.slots: List[HistorySlot] = kwargs.get('slots', [])
        self._checkpoints: Dict[str, list[HistorySlot]] = kwargs.get('_checkpoints', {})
        self.llm_model_name = llm_model_name

    def add_slot(self, history_slot: HistorySlot, position: int | SlotPosition = SlotPosition.BOTTOM) -> None:
        """
        Adds a new slot to the history at the specified position.

        Parameters
        ----------
        history_slot : HistorySlot
            The slot to add to the history.
        position : int | SlotPosition, optional
            The position where to add the slot. Can be an integer or a SlotPosition enum value.
            Defaults to SlotPosition.BOTTOM.
        """
        if isinstance(position, SlotPosition):
            if position == SlotPosition.BOTTOM:
                self.slots.append(history_slot)
            elif position == SlotPosition.TOP:
                self.slots.insert(0, history_slot)
        else:
            self.slots.insert(position, history_slot)

    def get_last_slot(self) -> HistorySlot:
        """
        Returns the last slot of the history. A good syntactic sugar to get the last item from the conversation.

        Returns
        -------
        HistorySlot
            The last slot in the history.

        Raises
        ------
        IndexError
            If the history is empty.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        return self.slots[-1]

    def get_slot_by_index(self, index: int) -> HistorySlot:
        """
        Returns the slot at the given index.

        Parameters
        ----------
        index : int
            The index of the slot to return.

        Returns
        -------
        HistorySlot
            The slot at the given index.

        Raises
        ------
        IndexError
            If the history is empty or the index is out of range.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        return self.slots[index]

    def get_slot_by_id(self, id: str) -> HistorySlot:
        """
        Returns the slot with the given ID.

        Parameters
        ----------
        id : str
            The ID of the slot to return.

        Returns
        -------
        HistorySlot
            The slot with the given ID.

        Raises
        ------
        IndexError
            If the history is empty.
        ValueError
            If no slot with the given ID is found.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        for slot in self.slots:
            if slot.id == id:
                return slot
        raise ValueError(f"Slot with id {id} not found in history.")

    def get_slot_by_message(self, message: GenericMessage) -> HistorySlot:
        """
        Returns the slot containing the given message.

        Parameters
        ----------
        message : GenericMessage
            The message to search for.

        Returns
        -------
        HistorySlot
            The slot containing the message.

        Raises
        ------
        IndexError
            If the history is empty.
        ValueError
            If no slot contains the given message.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        for slot in self.slots:
            if any(message.id == msg.id for msg in slot.get_all_messages()):
                return slot
        raise ValueError(f"Message with id {message.id} not found in history.")

    def add_message(self, message: GenericMessage) -> HistorySlot:
        """
        Adds a new message to the history by creating a new slot.

        Parameters
        ----------
        message : GenericMessage
            The message to add to the history.

        Returns
        -------
        HistorySlot
            The new slot containing the message added to the history. Useful for chaining.
        """
        slot = HistorySlot([message])
        self.slots.append(slot)
        return slot

    def delete_slot(self, slot: HistorySlot) -> None:
        """
        Deletes a slot from the history.

        Parameters
        ----------
        slot : HistorySlot
            The slot to delete.

        Raises
        ------
        ValueError
            If the slot is not found in the history.
        """
        if slot not in self.slots:
            raise ValueError("Slot not found in history.")

        self.slots.remove(slot)
        logging.debug(f"Slot {slot} deleted from history.")

    def delete_slot_by_id(self, slot_id: str) -> None:
        """
        Deletes a slot from the history by its ID. If the ID does not exist, it logs a warning.

        Parameters
        ----------
        slot_id : str
            The ID of the slot to delete.
        """
        slot_to_delete = next((slot for slot in self.slots if slot.id == slot_id), None)
        if not slot_to_delete:
            logging.warning(f"No slot found with ID {slot_id}.")
        else:
            self.slots.remove(slot_to_delete)
            logging.debug(f"Slot with ID {slot_id} deleted from history.")

    def delete_message(self, message: Message) -> None:
        """
        Deletes a message from all slots in the history.

        Parameters
        ----------
        message : Message
            The message to delete.

        Raises
        ------
        ValueError
            If the message is not found in any slot.
        """
        for slot in self.slots:
            if message in slot.messages:
                if len(slot.messages) == 1:
                    self.delete_slot(slot)
                    logging.debug(f"Slot {slot} deleted from history because it contained only the message to delete.")
                else:
                    slot._delete_message_by_id(message.id)
                    logging.debug(f"Message {message} deleted from slot {slot}.")
                return
        raise ValueError("Message not found in any slot.")

    def delete_message_by_id(self, message_id: str) -> None:
        """
        Deletes a message from all slots in the history by its ID.

        Parameters
        ----------
        message_id : str
            The ID of the message to delete. If the ID does not exist, it logs a warning.
        """
        for slot in self.slots:
            message_to_delete = next((message for message in slot.messages if message.id == message_id), None)
            if message_to_delete:
                if len(slot.messages) == 1:
                    self.delete_slot(slot)
                    logging.debug(f"Slot {slot} deleted from history because it contained only the message to delete.")
                else:
                    slot._delete_message_by_id(message_to_delete.id)
                    logging.debug(f"Message with ID {message_id} deleted from slot {slot}.")
                return
        logging.warning(f"No message found with ID {message_id}.")

    def _export(self) -> Dict:
        """
        Returns the history as a dictionary.
        Mainly used to export the object to a file.

        Returns
        -------
        Dict
            A dictionary representation of the history.
        """
        members_as_dict = self.__dict__.copy()
        members_as_dict["_checkpoints"] = {}

        # Exporting checkpoints
        for uid, slots in self._checkpoints.items():
            exported_slots = [slot._export() for slot in slots]
            members_as_dict["_checkpoints"][uid] = exported_slots

        # Exporting slots
        slots_list: List[Dict] = []
        for slot in self.slots:
            slots_list.append(slot._export())
        members_as_dict["slots"] = slots_list
        return members_as_dict

    @staticmethod
    def create_instance(members: Dict):
        """
        Creates a new instance of History from a dictionary.

        Parameters
        ----------
        members : Dict
            Dictionary containing the history data.

        Returns
        -------
        History
            A new instance of History.
        """
        # Loading slots
        members["slots"] = [HistorySlot.create_instance(slot) for slot in members["slots"]]

        # Loading checkpoints
        for uid, slots in members["_checkpoints"].items():
            members["_checkpoints"][uid] = [HistorySlot.create_instance(slot) for slot in slots]
        return History(**members)

    def get_messages_as_dict(self) -> List[Dict]:
        """
        Returns all messages in the history as a list of dictionaries.

        Returns
        -------
        List[Dict]
            List of message dictionaries.
        """
        formated_messages = []
        for slot in self.slots:
            formated_messages.append(slot.get_message().get_message_as_dict())
        return formated_messages

    def pretty_print(self) -> None:
        """
        Prints the history to stdout with colored output.
        """
        for slot in self.slots:
            message = slot.get_message()
            if message.role == MessageRole.USER:
                print('\033[92m[' + message.role.value + "]:\n" + message.get_as_pretty() + '\033[0m')
            elif message.role == MessageRole.ASSISTANT:
                print('\033[95m[' + message.role.value + "]:\n" + message.get_as_pretty() + '\033[0m')
            elif message.role == MessageRole.SYSTEM:
                print('\033[93m[' + message.role.value + "]:\n" + message.get_as_pretty() + '\033[0m')
            elif message.role == MessageRole.TOOL:
                print('\033[96m[' + message.role.value + "]:\n" + message.get_as_pretty() + '\033[0m')
            print("")

    def create_check_point(self) -> str:
        """
        Creates a checkpoint of the current history state.

        Returns
        -------
        str
            A unique identifier for the checkpoint.
        """
        uid: str = str(uuid.uuid4())
        self._checkpoints[uid] = copy.deepcopy(self.slots)
        return uid

    def load_check_point(self, uid: str) -> None:
        """
        Loads a checkpoint of the history. Perfect for a timey wimey rollback in time.

        Parameters
        ----------
        uid : str
            The unique identifier of the checkpoint to load.
        """
        self.slots = self._checkpoints[uid]

    def get_message(self, index) -> GenericMessage:
        """
        Returns the message at the given index.

        Parameters
        ----------
        index : int
            The index of the message to return.

        Returns
        -------
        GenericMessage
            The message at the given index.

        Raises
        ------
        IndexError
            If the history is empty or the index is out of range.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        return self.slots[index].get_message()

    def get_messages_by_tags(self, tags: List[str], strict=False) -> Sequence[GenericMessage]:
        """
        Returns messages that match the given tags based on the matching mode.

        Parameters
        ----------
        tags : List[str]
            The tags to filter messages by.
        strict : bool, optional
            Controls the matching mode:
            - If False (default), returns messages that have ANY of the specified tags.
              For example, searching for ["tag1"] will match messages with ["tag1", "tag2"].
              This is useful for broad filtering.
            - If True, returns messages that have EXACTLY the specified tags (and possibly more).
              For example, searching for ["tag1", "tag2"] will match messages with ["tag1", "tag2", "tag3"]
              but not messages with just ["tag1"] or ["tag2"].
              This is useful for precise filtering.

        Returns
        -------
        Sequence[GenericMessage]
            List of messages matching the tag criteria.

        Raises
        ------
        IndexError
            If the history is empty.

        Examples
        --------
        # Find all messages with tag1 (broad matching)
        `history.get_messages_by_tags(["tag1"])`
        # Returns messages with ["tag1"], ["tag1", "tag2"], etc.

        # Find messages with exactly tag1 and tag2 (strict matching)
        `history.get_messages_by_tags(["tag1", "tag2"], strict=True)`
        # Returns messages with ["tag1", "tag2"], ["tag1", "tag2", "tag3"], etc.
        # But not messages with just ["tag1"] or ["tag2"]
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        messages = []
        for slot in self.slots:
            message_tags = set(slot.get_message().tags)
            if strict is False:
                # Non-strict mode: message must have ANY of the specified tags
                if set(tags).intersection(message_tags):
                    messages.append(slot.get_message())
            else:
                # Strict mode: message must have ALL specified tags
                if set(tags).issubset(message_tags):
                    messages.append(slot.get_message())


        return messages

    def get_last_message(self) -> GenericMessage:
        """
        Returns the last message in the history.

        Returns
        -------
        GenericMessage
            The last message in the history.

        Raises
        ------
        IndexError
            If the history is empty.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        return self.slots[-1].get_message()

    def get_all_messages(self) -> List[Message]:
        """
        Returns all messages in the history.

        Returns
        -------
        List[Message]
            List of all messages in the history.

        Raises
        ------
        IndexError
            If the history is empty.
        """
        if len(self.slots) <= 0:
            raise IndexError("History is empty (no slots, so no messages)")
        return [slot.get_message() for slot in self.slots]

    def clean(self) -> None:
        """
        Resets the history, preserving only the initial system prompt if present.
        """
        if len(self.slots) > 0 and self.slots[0].get_message().role == MessageRole.SYSTEM:
            self.slots = [self.slots[0]]
        else:
            self.slots = []

    def _concat_history(self, history: Self) -> None:
        """
        Concatenates another history to this one.

        Parameters
        ----------
        history : Self
            The history to concatenate.
        """
        self.slots = self.slots + history.slots

    def get_token_count(self, hugging_face_repo_name: str | None = None, hugging_face_token: str | None = None, padding_per_message: int = 4, evaluate_all_history_as_one: bool = False) -> int:
        """
        Get the total token count of messages in the history.

        * If the hugging_face_repo_name is provided, the token count will be calculated using the transformers library.
        (If the llm is gated (private), the hugging_face_token must be provided to access the repo.)
        * If no Hugging Face details are given and the model is recognized as an OpenAi model then the token count will be calculated using the tiktoken library.
        * If none of the above conditions are met, an approximative token count will be returned. This is only a rough estimate and should not be used for precise calculations.

        Note that using Tiktoken and transformers are precise but quite slow (loging to HF using the token is the worst). The approximative token count is very fast but not precise.
        Important point: If you provide Hugging Face details, Yacana will try to evaluate the token count using the transformers library first even if the model is from OpenAi. If the Hugging Face evaluation fails then it won't fallback to Tiktoken but to the approximative method instead.

        Parameters
        ----------
        hugging_face_repo_name : str | None, optional
            The name of the Hugging Face repository for the model. If provided, the token count will be calculated using the transformers library.
        hugging_face_token : str | None, optional
            The Hugging Face access token for private models. Required if the model is gated (private).
        padding_per_message : int, optional
            The number of extra tokens to add per message for padding. Default is 4.
        evaluate_all_history_as_one : bool, optional
            If True, evaluates the entire history as a single input using the full chat template for token counting.
            This is more accurate but only available for Hugging Face models. Default is False.

        Returns
        -------
        int
            The total token count of all messages in history.
        """
        if evaluate_all_history_as_one is True and hugging_face_repo_name is None:
            raise IllogicalConfiguration("Parameter `evaluate_all_history_as_one` can only be used when hugging_face_repo_name is provided. It allows to count the tokens of the entire history at once using the full chat template instead of applying the template one message at a time which is less accurate. But it's only available for hugging face models that's why you need to set `hugging_face_repo_name`.")
        if hugging_face_token is not None and hugging_face_repo_name is None:
            raise IllogicalConfiguration("Parameter `hugging_face_token` can only be used when `hugging_face_repo_name` is provided. The token is used to access private models on Hugging Face Hub.")

        try:
            if hugging_face_repo_name and evaluate_all_history_as_one is True:
                logging.debug("Will try to count tokens using the whole history instead of message per message.")
                hf_messages_list: List[HFMessage] = [message.get_message_as_hugging_face_dict() for message in self.get_all_messages()]
                return count_tokens_using_huggingface(hf_messages_list, hugging_face_repo_name, hugging_face_token=hugging_face_token) + padding_per_message * len(hf_messages_list)
        except SpecializedTokenCountingError:
            logging.warning("Falling back to per-message token counting even though `evaluate_all_history_as_one` was True due to error in hugging_face token counting.")

        tokens = 0
        for message in self.get_all_messages():
            tokens += message.get_token_count(llm_model_name=self.llm_model_name, hugging_face_repo_name=hugging_face_repo_name, hugging_face_token=hugging_face_token, padding_per_message=padding_per_message)
        return tokens

    def __str__(self) -> str:
        """
        Returns a string representation of the history.

        Returns
        -------
        str
            A JSON string representation of the history.
        """
        result = []
        for slot in self.slots:
            result.append(slot.get_message()._export())
        return json.dumps(result)



import copy
import json
import logging
import requests
import uuid
from typing import Dict, List, Any, Optional

from .exceptions import McpBadToolConfig, McpBadTransport, McpResponseError, McpServerNotYetInitialized
from .tool import Tool, ToolType


class Mcp:
    """
    Connects to an HTTP streamable MCP server to get the tools and make them available to your Tasks.

    Parameters
    ----------
    server_url: str
        Url of the remote MCP server.
    headers: dict
        Dict of headers passed to all requests made to the MCP server. Use this for authentification.

    Attributes
    ----------
    server_url: str
        Url of the remote MCP server.
    headers: dict
        Dict of headers passed to all requests made to the MCP server. Use this for authentification.
    session: request.session
        The network session from connecting to the server using request.
    tools: List[Tool]
        The list of tools auto discovered on the server.
    initialized: bool
        Is the connection to the server established.
    session_id: Optional[str]
        Id of the request session.
    """

    def __init__(self, server_url: str, headers: dict = None) -> None:
        self.server_url = server_url.rstrip('/')
        self.headers = headers if headers is not None else {}
        self.session = requests.Session()
        self.tools: List[Tool] = []
        self.initialized = False
        self.session_id: Optional[str] = None

    def _make_request(self, method: str, params: Dict[str, Any] = None) -> Dict[str, Any]:
        """
        Makes a JSON-RPC request to the MCP server and parses the response as JSON or SSE (Server-Sent Events).

        Parameters
        ----------
        method : str
            The JSON-RPC method to call.
        params : Dict[str, Any], optional
            The parameters to pass to the method. Defaults to None.

        Returns
        -------
        Dict[str, Any]
            The result of the method call.
        """
        request_id = str(uuid.uuid4())
        payload: Dict[str, Any] = {
            "jsonrpc": "2.0",
            "id": request_id,
            "method": method
        }
        if params:
            payload["params"] = params

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json, text/event-stream",
            **self.headers
        }

        # Add session ID header if we have one
        if self.session_id:
            headers["Mcp-Session-Id"] = self.session_id

        try:
            response = self.session.post(
                self.server_url,
                json=payload,
                headers=headers,
                timeout=30
            )
            response.raise_for_status()

            # Check if server returned a session ID during initialization
            if method == "initialize" and "Mcp-Session-Id" in response.headers:
                self.session_id = response.headers["Mcp-Session-Id"]
                logging.debug(f"Server assigned session ID: {self.session_id}")

            # Handle different response types
            content_type = response.headers.get('content-type', '').lower()

            if 'application/json' in content_type:
                result = response.json()
                logging.debug(f"Getting result from MCP server with protocol: application/json. Response = {result}")
                if "error" in result:
                    raise McpResponseError(f"MCP Error: {result['error']}")
                return result.get("result", {})

            elif 'text/event-stream' in content_type:
                # Handle SSE response - for simplicity, we'll read the first JSON response
                # In a production client, you'd want to properly handle the SSE stream
                logging.debug(f"Getting result from MCP server with protocol: SSE. Response = {response}")
                return self._handle_sse_response(response)

            else:
                # If no specific content type, try to parse as JSON
                try:
                    logging.warning("Received unexpected content type, trying to parse as JSON")
                    result = response.json()
                    logging.debug(f"Getting result from MCP server with unknown protocol. Response = {result}")
                    if "error" in result:
                        logging.error(f"MCP Error: {result['error']}")
                        raise McpResponseError(f"MCP Error: {result['error']}")
                    return result.get("result", {})
                except:
                    raise McpBadTransport(f"Unexpected response format: {response.text}")

        except requests.exceptions.HTTPError as e:
            raise McpBadTransport(f"HTTP Error {e.response.status_code}: {e.response.text}")

    def _handle_sse_response(self, response) -> Dict[str, Any]:
        """
        Handle Server-Sent Events (SSE) response.
        Reads the streamed response line by line and reconstructs SSE events properly.

        Parameters
        ----------
        response : requests.Response
            Streamed response from the server (with text/event-stream content-type).

        Returns
        -------
        Dict[str, Any]
            The first valid 'result' found in the SSE stream.
        """
        event_data_lines = []

        for line in response.iter_lines(decode_unicode=True):
            if line is None:
                continue

            line = line.strip()
            # Comment line or empty
            if line == '':
                # End of one event
                if event_data_lines:
                    full_data = '\n'.join(event_data_lines)
                    try:
                        data = json.loads(full_data)
                        if "result" in data:
                            #print("=> ", data["result"])
                            return data["result"]
                        elif "error" in data:
                            raise McpResponseError(f"MCP Error: {data['error']}")
                    except json.JSONDecodeError:
                        pass  # Ignore and continue
                    event_data_lines = []  # Reset buffer for next event
                continue

            if line.startswith('data:'):
                #print("strip => ", line[5:].lstrip())
                # @todo Could be just 'data:' (empty), so use slicing carefully
                event_data_lines.append(line[5:].lstrip())

        raise McpBadTransport("Invalid SSE stream format")

    def connect(self) -> bool:
        """
        Initializes connection with the MCP server
        """
        logging.info(f"[MCP] Connecting to MCP server ({self.server_url})...")
        # Initialize the protocol
        init_result = self._make_request("initialize", {
            "protocolVersion": "2025-03-26",
            "capabilities": {
                "tools": {}
            },
            "clientInfo": {
                "name": "yacana-mcp-client",
                "version": "0.1.0"
            }
        })

        server_info = init_result.get('serverInfo', {})
        logging.info(f"[MCP] Connected to MCP server: {server_info.get('name', 'Unknown')} v{server_info.get('version', 'Unknown')}")

        # List available tools
        tools_result = self._make_request("tools/list")
        tools = tools_result.get("tools", [])

        for tool_info in tools:
            try:
                tool = Tool(tool_info.get("name"),
                            tool_info.get("description"),
                            function_ref=self._call_tool,
                            mcp_input_schema=tool_info["inputSchema"],
                            optional=True)
                self.tools.append(tool)
                logging.info(f"[MCP] Available tool: {tool.tool_name} - {tool.function_description}")
            except McpBadToolConfig as e:
                logging.warning(f"[MCP] Tool will be excluded due to : {e.message}")

        if not tools:
            logging.info("[MCP] No tools available on this server")

        self.initialized = True
        return True

    def _call_tool(self, tool_name: str, arguments: Dict[str, Any]) -> Dict[str, Any]:
        """
        Calls a tool on the MCP server using request.

        Parameters
        ----------
        tool_name : str
            The name of the tool to call.
        arguments : Dict[str, Any]
            The arguments to pass to the tool.

        Returns
        -------
        Dict[str, Any]
            The result of the tool call.
        """
        if not self.initialized:
            raise McpServerNotYetInitialized("Cannot call tool on server. Client not initialized")
        result = self._make_request("tools/call", {
            "name": tool_name,
            "arguments": arguments
        })

        return result

    def get_tools_as(self, tools_type: ToolType, optional=None) -> List[Tool]:
        """
        Returns the tools from the remote MCP server as a list of Tool objects. You choose how these tools will be called
        by specifying the tools_type parameter. This list of tools must be given to a Task() object.
        !Warning!: This returns a deep copy of the tools list !

        Parameters
        ----------
        tools_type : ToolType
            The type of tools to return.
        optional : bool, optional
            Changes the optional status of all tools in the MCP client. !WARNING! The status will remain for the next Tasks!

        Returns
        -------
        List[Tool]
            A list of Tool objects of the specified type.
        """

        if not self.initialized:
            raise McpServerNotYetInitialized("Cannot get tools from server. Client not initialized")
        tools_copy = copy.deepcopy(self.tools)
        for tool in tools_copy:
            tool.tool_type = tools_type
            if optional is not None:
                tool.optional = optional
        return tools_copy

    def forget_tool(self, tool_name: str) -> None:
        """
        Forgets about a tool from the MCP server. Note that using a tool leaves a trace in the history, hence it is not
        possible to completely forget about a tool. However, the tool won't be proposed for future tasks. Call this method
        right after connecting to the MCP server.

        Parameters
        ----------
        tool_name : str
            The name of the tool to delete.
        """
        if not self.initialized:
            raise McpServerNotYetInitialized("Cannot forget tool. Client not initialized")
        self.tools = [tool for tool in self.tools if tool.tool_name != tool_name]
        logging.info(f"Tool '{tool_name}' deleted from MCP client.")

    def disconnect(self):
        """
        Explicitly disconnects from the server
        """
        if self.session_id:
            try:
                headers = {"Mcp-Session-Id": self.session_id, **self.headers}
                self.session.delete(self.server_url, headers=headers, timeout=5)
            except:
                pass  # Ignore errors during cleanup
        self.session.close()
